/***************************************************************************
 *   Copyright (C) 2005 by John Bennett   *
 *   n4xi@arrl.net   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

//
// ax25.cpp
//

#define _AX25_CPP_

#ifdef HAVE_NETAX25_AX25_H
    #include <netax25/ax25.h>
#else
    #include <netax25/kernel_ax25.h>
#endif

#include <netax25/axlib.h>
#include <netax25/axconfig.h>

#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <net/if.h>
#include <net/ethernet.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "ax25.h"
#include "makeuc.h"
#include "strxcpy.h"
#include "logmsg.h"


static char szFrameError[] = "?-Frame Length Error";
#define FRAME_ERROR_SIZE    sizeof( szFrameError )


//////////////////////////////////////////////////////////////////////////////
// Constructor

Ax25::Ax25()
{
    *m_szPort = 0;
    m_socket = 0;
    m_nDigiIndex = 0;
    m_nDigiCnt = 0;
}


//////////////////////////////////////////////////////////////////////////////
// Destructor

Ax25::~Ax25()
{
    if( m_socket )
        close( m_socket );
}



//////////////////////////////////////////////////////////////////////////////
// setup

BOOL Ax25::setup( char * pMyCall, char * szPort, BOOL bRcvAll )
{
    uint16_t uFilter;
    char Call[AX25_CALLSSID_SIZE+1];

    if( *pMyCall == 0 || pMyCall == NULL )
        return FALSE;

    memset( Call, 0, AX25_CALLSSID_SIZE );
    strxcpy( Call, pMyCall, AX25_CALLSSID_SIZE );
    EncodeCall( Call );
    memcpy( m_MyCall, Call, ADDR_SIZE );

    m_TxPid = (unsigned char) NO_L3;
    m_TxCf = (unsigned char) UI;

    m_nUiFrameCount = 0;

    strxcpy( m_szPort, szPort, AXPORT_SIZE );

    if ( bRcvAll )
        uFilter = ETH_P_ALL;
    else
        uFilter = ETH_P_AX25;

    if ( (m_socket = socket(AF_PACKET, SOCK_PACKET, htons(uFilter))) == -1 )
    {
        g_log.PostMsg( APRMSG_AX25_SOCK_SETUP_ERR, NULL );
        return FALSE;
    }

    return TRUE;
}


//////////////////////////////////////////////////////////////////////////////
// GetUiFrameCount

int Ax25::UiFrameCount( void )
{
    int nCount;

    nCount = m_nUiFrameCount;
    m_nUiFrameCount = 0;

    return nCount;
}


//////////////////////////////////////////////////////////////////////////////
// put - output KISS frame

void Ax25::put( char * pCall, char * pDigi, char * pInf, int nInfLen )
{
    int nAddrLen;
    int nFrameLength;

    nAddrLen = SetAddr( pCall, pDigi );
    nFrameLength = AX25_FLAG_SIZE + nAddrLen + CF_SIZE + PID_SIZE + nInfLen;

    m_TxKissFrame[0] = 0; // AX25_FLAG will be written here;

    memcpy( AX25_FLAG_SIZE+m_TxKissFrame, m_AddrField, nAddrLen );
    m_TxKissFrame[AX25_FLAG_SIZE+nAddrLen] = m_TxCf;
    m_TxKissFrame[AX25_FLAG_SIZE+nAddrLen+CF_SIZE] = m_TxPid;
    if( nFrameLength > AX25_FRAME_MAX_SIZE )
        memcpy( &m_TxKissFrame[AX25_FLAG_SIZE+nAddrLen+CF_SIZE+PID_SIZE], szFrameError, FRAME_ERROR_SIZE );
    else
        memcpy( &m_TxKissFrame[AX25_FLAG_SIZE+nAddrLen+CF_SIZE+PID_SIZE], pInf, nInfLen );
    strcpy( m_sa.sa_data, m_szPort );

    if( sendto( m_socket, m_TxKissFrame, nFrameLength, 0, &m_sa, sizeof( m_sa ) ) == -1)
        g_log.PostMsg( APRMSG_AX25_SOCK_SENDTO_ERR, NULL );
}


//////////////////////////////////////////////////////////////////////////////
// get - receive and decode ax25 frame if it has arrived

BOOL Ax25::get( void )
{
    int i;
    struct pollfd Poll;
    char * pFrame;
    socklen_t nAddrLen;

    Poll.fd = m_socket;
    Poll.events = POLLIN | POLLPRI;

    if( poll( &Poll, 1, 1000 ) <= 0 )
        return FALSE;

    if( ((Poll.revents | POLLIN) || (Poll.revents | POLLPRI)) == 0 )
        return FALSE;

    memset( &m_RxKissFrame, 0, AX25_FRAME_MAX_SIZE );

    nAddrLen = sizeof( m_sa );
    m_nFrameLength = recvfrom( m_socket, m_RxKissFrame, AX25_FRAME_MAX_SIZE, 0, &m_sa, &nAddrLen );
    if( m_nFrameLength == -1 )
    {
        g_log.PostMsg( APRMSG_AX25_SOCK_RCVFRM_ERR, NULL );
        return FALSE;
    }

    if( m_nFrameLength == 0 )
        return FALSE;

    pFrame = m_RxKissFrame + 1;     // point past flag

    // "to" call
    DecodeCall( pFrame, &m_ToCall );

    // "from" call
    pFrame += ADDR_SIZE;
    DecodeCall( pFrame, &m_FmCall );

    m_nDigiCnt = 0;
    memset( &m_Digi, 0, sizeof( m_Digi ) );
    for( i=0; i<MAX_DIGI; i++ )
    {
        if( pFrame[6] & 0x01 )
            break;
        pFrame += ADDR_SIZE;
        DecodeCall( pFrame, &m_Digi[i] );
        m_nDigiCnt++;
    }

    pFrame += ADDR_SIZE;    // now control field

    if( !DecodeFrameType( pFrame ) )
        return FALSE;

    pFrame++;       // now PID field

    switch( FrameType() )
    {
        case I_FRAME:
        case U_FRM_UI:
            m_RxPid = *pFrame++;    // now information field
            m_nRxDataLength = m_nFrameLength - ( (m_nDigiCnt+2)*ADDR_SIZE + PID_SIZE + CF_SIZE );
            m_nUiFrameCount++;
            break;

        default:    // all others
            m_nRxDataLength = m_nFrameLength - ( (m_nDigiCnt+2)*ADDR_SIZE + CF_SIZE );
            break;

    }

    m_pRxData = pFrame;

    return TRUE;
}


//////////////////////////////////////////////////////////////////////////////
// data - get data frame from previously received frame.  Returns size.

int Ax25::data( char * pData, int nMaxLen )
{
    if( m_nRxDataLength > nMaxLen )
    {
        memcpy( pData, m_pRxData, nMaxLen );
        return nMaxLen;
    }
    else
    {
        memcpy( pData, m_pRxData, m_nRxDataLength );
        return m_nRxDataLength;
    }
}


//////////////////////////////////////////////////////////////////////////////
// ToCall - return call of "to" field

void Ax25::ToCall( char * pCall )
{
    strcpy( pCall, m_ToCall.szCall );
}


//////////////////////////////////////////////////////////////////////////////
// FmCall - return call of "fm" field

void Ax25::FmCall( char * pCall )
{
    strcpy( pCall, m_FmCall.szCall );
}


//////////////////////////////////////////////////////////////////////////////
// ToSsid - return ssid of call "to" field

int Ax25::ToSsid( void ) { return m_ToCall.nSsid; }


//////////////////////////////////////////////////////////////////////////////
// FmSsid - return ssid of call "fm" field

int Ax25::FmSsid( void ) { return m_FmCall.nSsid; }


//////////////////////////////////////////////////////////////////////////////
// FmSsid - return ssid of call "fm" field

int Ax25::RxPid( void ) { return (int) m_RxPid; }


//////////////////////////////////////////////////////////////////////////////
// SetAddr - encode ax.25 address field.  Callsigns are UC and left shifted
// one bit.  The destination address comes first, and can be followed by up
// to 8 digis.  The last call in the sequence has the LSB set to mark the
// end of the address field.  Max field size is 70 bytes (560 bits). The digi
// calls must be in the comma delimited format.  Returns length of field.

int Ax25::SetAddr( char * pCall, char * pDigi )
{
    int nPos = ADDR_SIZE*2;
    char * pS;
    char szTemp[128];
    char Digi[AX25_CALLSSID_SIZE+1];
    int nDigiCount;

    memset( &m_AddrField, 0, ADDR_FIELD_SIZE );

    strcpy( szTemp, pCall );
    EncodeCall( szTemp );
    memcpy( m_AddrField, szTemp, ADDR_SIZE );
    memcpy( m_AddrField+ADDR_SIZE, m_MyCall, ADDR_SIZE );

    nDigiCount = 2;         // always destination + source

    if( pDigi )
    {
        strcpy( szTemp, pDigi );
        pS = strtok( szTemp, ", " );
        while( pS )
        {
            strxcpy( Digi, pS, AX25_CALLSSID_SIZE );
            EncodeCall( Digi );
            memcpy( m_AddrField+nPos, Digi, ADDR_SIZE );
            nPos += ADDR_SIZE;
            nDigiCount++;
            pS = strtok( NULL, ", " );
        }
    }

    // set LSB of last call to mark end of address field
    *(m_AddrField+(nPos-1)) = (char) (  *(m_AddrField+(nPos-1)) | 0x01 );

    // mark end of string
    *(m_AddrField+nPos) = 0;

    return nDigiCount * ADDR_SIZE;
}

//////////////////////////////////////////////////////////////////////////////
// EncodeCall - convert all characters to uc, pad trailing spaces if required,
// trim '-1', etc. Add SSID and set H and R bits as appropriate.  If no SSID
// specified, make default (zero).  It is assumed the string has the required
// length to accomodate padding if required.

void Ax25::EncodeCall( char * pCall )
{

    int nSSID;
    int nCount;
    char * pS;

    makeuc( pCall );       // insure uc for ax.25

    pS = strchr( pCall, '-' );
    if( pS )
    {
        *pS++ = 0;
        nSSID = atoi( pS );
    }
    else
        nSSID = 0;

    nCount = 0;
    pS = pCall;
    while( *pS )
    {
        nCount++;
        *pS = (char) ( *pS << 1 );
        pS++;
    }

    while( nCount < 6 )
    {
        *pS++ = AX25_SPACE;
        nCount++;
    }

    *pS++ = (char) ( (nSSID<<1) | 0x60 );

    *pS = 0;

}


//////////////////////////////////////////////////////////////////////////////
// DecodeCall - decode left-shifted call

void Ax25::DecodeCall( char * pCall, struct ax25call * pResult )
{
    unsigned char szCall[ADDR_SIZE+1];
    unsigned char * pC;
    int i;

    memcpy( szCall, pCall, ADDR_SIZE );
    pC = szCall;

    pResult->nSsid = (pCall[6] >> 1) & 0x0f;

    for( i=0; i<6; i++ )
    {
        if( *pC == 0x40 )
            break;
        *pC++ >>= 1;
    }

    *pC = 0;
    strcpy( pResult->szCall, (char *) szCall );
}


//////////////////////////////////////////////////////////////////////////////
// DecodeFrameType - determine if I, S, or U type frame. Return TRUE sucess,
// FALSE failure.
//

BOOL Ax25::DecodeFrameType( char * pByte )
{
    // the default
    m_nRxCf = NO_FRAME;

    if( (*pByte & 0x01) == 0 )
    {
        m_nRxCf = I_FRAME;
        return TRUE;
    }

    if( (*pByte & 0x03) == 0x01 )
    {
        switch( *pByte &0x03 )
        {
            case RR:
                m_nRxCf = S_FRM_RR;
                break;

            case RNR:
                m_nRxCf = S_FRM_RNR;
                break;

            case REJ:
                m_nRxCf = S_FRM_REJ;
                break;

            default:
                return FALSE;
        }

        return TRUE;

    }

    if( (*pByte & 0x03) == 0x03 )
    {
        switch( *pByte & 0xef )
        {
            case UI:
                m_nRxCf = U_FRM_UI;
                break;

            case FRMR:
                m_nRxCf = U_FRM_FRMR;
                break;

            case UA:
                m_nRxCf = U_FRM_UA;
                break;

            case DM:
                m_nRxCf = U_FRM_DM;
                break;

            case DISC:
                m_nRxCf = U_FRM_DISC;
                break;

            case SABM:
                m_nRxCf = U_FRM_SABM;
                break;

            default:
                return FALSE;

            }

        return TRUE;

        }

    return FALSE;
}


//////////////////////////////////////////////////////////////////////////////
// FrameType - returns type of frame for current packet
//

int Ax25::FrameType( void )
{
    return m_nRxCf;
}


int Ax25::InfLength( void )
{
    return m_nRxDataLength;
}


#ifdef _RAWFRAME

char * Ax25::RawFrame( void )
{
    return m_RxKissFrame;

}

unsigned Ax25::RawFrameLength( void )
{
    return m_nFrameLength;
}

#endif

//////////////////////////////////////////////////////////////////////////////
// Digipeater access routines
//

int Ax25::DigiCount( void )
{
    return m_nDigiCnt;
}


BOOL Ax25::DigiIndex( int nAction )
{
    switch( nAction )
    {
        case INDEX_RESET:
            m_nDigiIndex = 0;
            break;

        case INDEX_NEXT:
            if( (m_nDigiIndex+1) >= m_nDigiCnt )

                return FALSE;
            m_nDigiIndex++;
            break;

        case INDEX_PREV:
            if( m_nDigiIndex == 0 )
                return FALSE;
            m_nDigiIndex--;
            break;

        default:
            return FALSE;
    }

    return TRUE;

}

//////////////////////////////////////////////////////////////////////////////
// DigiCall - return call of digipeater in the current index

void Ax25::DigiCall( char * pCall )
{
    sprintf( pCall, "%s-%i", m_Digi[m_nDigiIndex].szCall, m_Digi[m_nDigiIndex].nSsid );
}

