/***************************************************************************
 *   Copyright (C) 2005 by John Bennett   *
 *   n4xi@arrl.net   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

//
// wxnaprd - the ax25 tx/rx daemon for wxn server suite.
//

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pthread.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <netax25/daemon.h>
#include <time.h>
#include <syslog.h>

// libwxn
#include <logmsg.h>
#include <tz.h>
#include <cconfig.h>

#include "qrz.h"
#include "callbook.h"
#include "aprs.h"
#include "aprsproc.h"
#include "aconfig.h"
#include "node.h"

static void signal_handler(int sig);
void * RxAprsThread( void * pArg );
void * TxAprsThread( void * pArg );
static void ExitFail( int nMsg, char * sInfo );

// global data

qrz g_Qrz;
Callbook g_callbook;
Aprs g_aprs;
noderx g_Local;
noderx g_RxNode;
BOOL g_bVerbose = FALSE;
logmsg g_log;

// locals @ file scope

static BOOL g_bStop = FALSE;
static pthread_t g_hTxThread;
static pthread_t g_hRxThread;


static void signal_handler(int sig)
{
    switch (sig)
    {
        case SIGTERM:
        case SIGQUIT:
            g_bStop = TRUE;
            break;

        default:
            break;
    }
}


static void ExitFail( int nMsg, char * sInfo )
{
    g_log.PostMsg( nMsg, sInfo );
    syslog( LOG_ERR, "Terminating - check app log" );
    exit( 1 );
}



int main(int argc, char *argv[])
{
    BOOL bStartDaemon = TRUE;
    timespec ts;
    int nStatus;

    if( argc > 1 )
    {
        if( strcmp( argv[1], "--debug" ) == 0 )
        {
            g_bVerbose = TRUE;
            bStartDaemon = FALSE;
        }
        else if( strcmp(argv[1], "--verbose") == 0 )
            g_bVerbose = TRUE;
        else if( strcmp(argv[1], "--build") == 0 )
        {
            fprintf( stdout, "wxnaprd: ver %s (built %s @ %s)\n", VERSION, __DATE__, __TIME__ );
            exit( 0 );
        }
    }

    g_log.init( "/var/log/wxn/wxnaprd.log", g_bVerbose );

    if( bStartDaemon )
    {
        if( !daemon_start(TRUE) )
            ExitFail( MSG_DAEMON_START_FAIL, NULL );
    }

    // set signal handlers
    signal(SIGTERM, signal_handler);
    signal(SIGHUP,  InitLists);
    signal(SIGQUIT, signal_handler);
    signal(SIGPIPE, SIG_IGN);

    // print signon message
    if( g_bVerbose )
    {
        fprintf( stdout, "wxnaprd: AX.25 daemon for Linux, ver %s (built %s @ %s)\n", VERSION, __DATE__, __TIME__ );
        fprintf( stdout, "Process ID: %i\n", (int) getpid() );
    }

    g_log.PostMsg( MSG_INIT_BEGIN, NULL );
    syslog( LOG_INFO, "started" );

    //////////////////////// setup time variables
    g_Time.Init();
    g_Time.Set();

    //////////////////////// read config files
    if( !ReadComConfig() )
        ExitFail( MSG_STARTUP_FAIL, NULL );

    if( !ReadAprConfig() )
        ExitFail( MSG_STARTUP_FAIL, NULL );

    //////////////////////// setup timespec struct for nanosleep
    ts.tv_sec = 0;
    ts.tv_nsec = 100000000; // 0.1 seconds

    //////////////////////// callbook and qrz are mutually exclusive
    if( !g_AprConfig.bUseCallbook &&  g_AprConfig.bUseQrz )
    {
        if( !g_Qrz.Init( g_AprConfig.szQrzPath ) )
            g_log.PostMsg( APRMSG_QRZ_INIT_FAIL, NULL );
    }

    //////////////////////// aprs setup
    g_aprs.setup( g_ComConfig.szServerCall,
                  g_ComConfig.szAprsToCall,
                  g_ComConfig.szAprsDigis,
                  g_ComConfig.szAx25Port,
                  g_AprConfig.szAprsText,
                  g_AprConfig.bRcvLocal );
    g_aprs.SetCoordinates( g_ComConfig.dLatitude, g_ComConfig.dLongitude );
    AprsStnTypeInit();
    AprsRandomTimerInit();


    //////////////////////// start tx thread
    nStatus = pthread_create( &g_hTxThread, NULL, &TxAprsThread, NULL );
    if( nStatus )
        ExitFail( MSG_THREAD_START_FAIL, "Tx" );

    //////////////////////// init 'deny' list
    InitLists( 0 );

    //////////////////////// if configured, start rx thread
    if( g_AprConfig.bAprsReceive )
    {
        nStatus = pthread_create( &g_hRxThread, NULL, &RxAprsThread, NULL );
        if( nStatus )
            g_log.PostMsg( MSG_THREAD_START_FAIL, "Rx" );
    }

    g_log.PostMsg( MSG_STARTUP_OK, NULL );

    // time loop
    FOREVER
    {
        if( g_bStop )
            break;

        g_Time.Set();

        nanosleep( &ts, NULL );
    }

    pthread_join( g_hRxThread, NULL );
    pthread_join( g_hTxThread, NULL );

    g_log.PostMsg( MSG_DAEMON_STOP, NULL );
    syslog( LOG_INFO, "stopped" );

    return EXIT_SUCCESS;

}


////////////////////////////////////////////////////////////////////
// APRS receive loop
//

void * RxAprsThread( void * pArg )
{
    //FOREVER
    for(;;)
    {
        switch( g_aprs.rxStatus() )
        {
            case NO_DATA:
                break;

            case WX_DATA:
                AprsReceive();
                break;

            case POSITION:
                AprsPosition();
                break;

            case WXN_DATA:
                AprsWxnData();
                break;

            default:
                break;
        }

        if( g_bStop )
            break;
    }

    return NULL;
}


//////////////////////////////////////////////////////////////////////////////
// APRS Tx loop (APRS beacons)
//

void * TxAprsThread( void * pArg )
{
    BOOL bNoEntry = FALSE;
    BOOL bReadStatus = FALSE;

    //FOREVER
    for(;;)
    {
        if( (g_Time.Minute()%HOST_INTERVAL == g_nAprsMinute) && (g_Time.Second() >= g_nAprsSecond) )
        {
            if( !bNoEntry )
            {
                bNoEntry = TRUE;
                bzero( &g_Local, sizeof( g_Local ) );
                strcpy( g_Local.Info.szCallSsid, g_ComConfig.szServerCall );
                bReadStatus = ReadNode( &g_Local );
                if( !bReadStatus )
                    g_log.PostMsg( APRMSG_BEACON_DATA_READ_FAIL, NULL );
                else
                {
                    g_aprs.beacon( &g_Local );

                    if( g_bStop )
                        break;

                    if( ((g_nExtraTimeOffset - g_Time.Minute()) % g_AprConfig.nAprsExtraInterval) == 0 )
                    {
                        sleep( AprsRandomDelay() );
                        g_aprs.BeaconExtra();
                        if( g_bStop )
                            break;

                        sleep( AprsRandomDelay() );
                        g_aprs.WxnTx( TYPE_A, &g_Local );
                        if( g_bStop )
                            break;

                        sleep( AprsRandomDelay() );
                        g_aprs.WxnTx( TYPE_B, &g_Local );
                        if( g_bStop )
                            break;

                        sleep( AprsRandomDelay() );
                        g_aprs.WxnTx( TYPE_C, &g_Local );
                        if( g_bStop )
                            break;

                        sleep( AprsRandomDelay() );
                        g_aprs.WxnTx( TYPE_D, &g_Local );
                        if( g_bStop )
                            break;

                        // only tx if data exists
                        if( g_Local.Data.rmax )
                        {
                            sleep( AprsRandomDelay() );
                            g_aprs.WxnTx( TYPE_E, &g_Local );
                        }
                    }
                }
            }
        }

        else
            bNoEntry = FALSE;

        if( g_bStop )
            break;

        sleep( 1 );
    }

    return NULL;
}
