/***************************************************************************
 *   Copyright (C) 2005 by John Bennett   *
 *   n4xi@arrl.net   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

//
// dav2.cpp - driver for Davis Monitor II and compatible weather stations
//

#define _DAV2_CPP_

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <pthread.h>
#include <unistd.h>

// libwxn
#include <logmsg.h>

#include "precip.h"
#include "barom.h"
#include "humid.h"
#include "temp.h"
#include "wind.h"
#include "wxport.h"
#include "sensors.h"
#include "rates.h"
#include "hglobal.h"
#include "fixvalue.h"
#include "wxdata.h"
#include "dav2.h"


#ifndef byte_t
    typedef unsigned char byte_t;
#endif

#ifndef word_t
    typedef unsigned short word_t;
#endif

#define ACK         0x06
#define NACK        '!'
#define DAV_ERROR   0x80

#define BLOCK_START 0x01

#define ACK_SIZE        1
#define BLOCK_SIZE      18
#define BLOCKS          40      // allow some headroom
#define DATA_ARRAY_SIZE ( BLOCK_SIZE * BLOCKS )

// Offsets within the block

#define SOB     0
#define TIN     1
#define TOUT    3
#define WSPD    5
#define WDIR    6
#define BAR     8
#define HIN     10
#define HOUT    11
#define RAIN    12
#define CRC     16

#define MAX_DATA_SIZE   128


//////////////////////////////////////
// local function prototypes
//////////////////////////////////////

static BOOL GetAck( void );
static BOOL ResetLink( void );
static BOOL GetBlock( byte_t * pBlock );

// thread related functions

static BOOL Dav2StartThread( BOOL bVerbose );
static void * Dav2Thread( void * pParam );


//////////////////////////////////////
// global variables @ file scope

// temporary values
static int g_nSpeedAccum = 0;
static int g_nDirAccum = 0;
static int g_nWindCount = 0;
static int g_nDav2Q1 = 0;
static int g_nDav2Q4 = 0;

static int g_nGustSpeed = 0;
static int g_nGustDir = 0;

static int g_nHiTemp;
static int g_nLoTemp;

// flags
static BOOL g_bTempRead = FALSE;

//////////////////////////////////////

pthread_mutex_t mx_Dav2Lock = PTHREAD_MUTEX_INITIALIZER;



// worker thread
static pthread_t g_hThread;
static BOOL g_bStopThread = FALSE;

////////////////////////////////////////////////
// Davis commands
//

//static char g_szStopCmd[] = "STOP\r";
//static char g_szStartCmd[] = "START\r";

static char g_szLoopDataCmd[] = "LOOP\xff\xff\r";
#define LOOPDATA_CMD_LEN sizeof( g_szLoopDataCmd ) - 1

//////////////////////////////////////////////////////////
// Calibration/sensor detect commands/variables

static int g_nCalTempIn;
static char g_szInTempCalCmd[] = "WRD\x44\x52\r";
#define INTEMPCAL_CMD_LEN sizeof( g_szInTempCalCmd ) - 1

static int g_nCalTempOut;
static char g_szOutTempCalCmd[] = "WRD\x44\x78\r";
#define OUTTEMPCAL_CMD_LEN sizeof( g_szOutTempCalCmd ) - 1

static int g_nCalBarom;
static char g_szBaromCalCmd[] = "WRD\x44\x2c\r";
#define BAROMCAL_CMD_LEN sizeof( g_szBaromCalCmd ) - 1

static int g_nCalHumidOut;
static char g_szOutHumCalCmd[] = "WRD\x44\xda\r";
#define OUTHUMCAL_CMD_LEN sizeof( g_szOutHumCalCmd ) - 1


//////////////////////////////////////////////////////////
// Data commands

static char g_szInHumidCmd[] = "WRD\x24\x80\r";
#define INHUMID_CMD_LEN sizeof( g_szInHumidCmd ) - 1

static char g_szOutHumidCmd[] = "WRD\x24\x98\r";
#define OUTHUMID_CMD_LEN sizeof( g_szOutHumidCmd ) - 1



//////////////////////////////////////////////////////////////////////////////
// Dav2ProcessData - read weather station, set hi/low values, etc. Returns
// TRUE success, FALSE on failure.
//

BOOL Dav2ProcessData( void )
{

    byte_t DataBlock[BLOCK_SIZE];
    int nVal;
    int nAttempts;
    cwinds Wind;

    pthread_mutex_lock( &mx_Dav2Lock );

    ////////////////////////// get data packet from station

    nAttempts = 2;
    while( !GetBlock(DataBlock) )
    {
        nAttempts--;
        if( nAttempts == 0 )
        {
            g_log.PostMsg( HOSTMSG_GETBLOCK_FAIL, NULL );
            pthread_mutex_unlock( &mx_Dav2Lock );
            return FALSE;
        }

        ResetLink();

    }

    ////////////////////////////////// TEMPERATURE

    nVal = (short int) (DataBlock[TOUT] | (DataBlock[TOUT+1] << 8));
    nVal = FixValue( nVal + g_nCalTempOut );

    if( !g_bTempRead )
    {
        g_nHiTemp = nVal;
        g_nLoTemp = nVal;
    }

    TempSet( nVal, g_nHiTemp, g_nLoTemp );

    // reset hi/lo temps
    g_nHiTemp = nVal;
    g_nLoTemp = nVal;
    g_bTempRead = FALSE;

    // inside

    nVal = (short int) (DataBlock[TIN] | (DataBlock[TIN+1] << 8));
    nVal = FixValue( nVal + g_nCalTempIn );
    InsideTempSet( nVal );

    ////////////////////////////////// Barometric Pressure

    nVal = (short int) (DataBlock[BAR] | (DataBlock[BAR+1] << 8));
    nVal = FixValue( nVal- g_nCalBarom );
    nVal = MillibarsT( nVal );
    BaromSet( nVal );

    ////////////////////////////////// Humidity

    if( g_Sensors.bOutsideHumid )
    {
        nVal = (short int) DataBlock[HOUT];
        nVal += g_nCalHumidOut;
        if( nVal <= 100 )
            HumidSet( nVal );
    }

    if( g_Sensors.bInsideHumid )
    {
        nVal = (short int) DataBlock[HIN];
        if( nVal <= 100 )
            InsideHumidSet( nVal );
    }

    ////////////////////////////////// WIND

    // add current to accums
    g_nSpeedAccum += (short int) DataBlock[WSPD];
    g_nDirAccum += (short int) (DataBlock[WDIR] | (DataBlock[WDIR+1] << 8));
    g_nWindCount++;

    //// averages
    Wind.nSpeed = g_nSpeedAccum / g_nWindCount;

    if( g_nDav2Q1 && g_nDav2Q4 )
    {
        g_nDirAccum += g_nDav2Q1 * 360;
    }
    Wind.nDir = g_nDirAccum / g_nWindCount;
    if( Wind.nDir > 359 )
        Wind.nDir -= 360;

    //// wind gusts
    if( g_nWindCount == 1 )
    {
        Wind.nGustSpeed = Wind.nSpeed;
        Wind.nGustDir = Wind.nDir;
    }
    else if( g_nWindCount > 1 )
    {
        Wind.nGustSpeed = g_nGustSpeed;
        Wind.nGustDir = g_nGustDir;
    }

    // reset gust speed/dir/accum
    g_nGustSpeed = 0;
    g_nGustDir = 0;
    g_nSpeedAccum = 0;
    g_nDirAccum = 0;
    g_nWindCount = 0;
    g_nDav2Q1 = 0;
    g_nDav2Q4 = 0;

    WindSet( &Wind );

    ////////////////////////////////// Precip

    nVal = (short int) (DataBlock[RAIN] | (DataBlock[RAIN+1] << 8));
    PrecipSet( nVal );

    ////////////////////////////////// Clean-Up

    pthread_mutex_unlock( &mx_Dav2Lock );

    return TRUE;

}


//////////////////////////////////////////////////////////////////////////////
// GetBlock - retrieves block of data (packet) from Davis wx stn. Returns
// TRUE success, FALSE failure.
//

BOOL GetBlock( byte_t * pBlock )
{

    if( !WriteWxStn( g_szLoopDataCmd, LOOPDATA_CMD_LEN ) )
        return FALSE;
    if( !GetAck() )
        return FALSE;
    if( !ReadWxStnBytes( pBlock, BLOCK_SIZE ) )
    {
        g_log.PostMsg( HOSTMSG_READWXSTNBYTES_FAIL, NULL );
        return FALSE;
    }

    if( *pBlock != BLOCK_START )
        return FALSE;
    else
        return TRUE;

}


//////////////////////////////////////////////////////////////////////////////
// Dav2Init - check connection, determine attached sensors.
//

BOOL Dav2Init( BOOL bVerbose )
{

    byte_t Resp[2];
    byte_t DataBlock[BLOCK_SIZE];
    int nVal;
    int nAttempts;

    //// resync
    if( !ResetLink() )
        return FALSE;

    //////////////////////// determine if humidity sensors attached

    //// Indoor

    if( !WriteWxStn( g_szInHumidCmd, INHUMID_CMD_LEN ) )
        return FALSE;
    if( !GetAck() )
        return FALSE;
    if( !ReadWxStnBytes( Resp, 1 ) )
    {
        g_log.PostMsg( HOSTMSG_READWXSTNBYTES_FAIL, NULL );
        return FALSE;
    }
    if( *Resp != DAV_ERROR )
        g_Sensors.bInsideHumid = TRUE;


    //// Outdoor

    if( !WriteWxStn( g_szOutHumidCmd, OUTHUMID_CMD_LEN ) )
        return FALSE;
    if( !GetAck() )
        return FALSE;
    if( !ReadWxStnBytes( Resp, 1 ) )
    {
        g_log.PostMsg( HOSTMSG_READWXSTNBYTES_FAIL, NULL );
        return FALSE;
    }
    if( *Resp != DAV_ERROR )
        g_Sensors.bOutsideHumid = TRUE;

    //////////////////////// read in calibration values

    //// Inside Temperature

    if( !WriteWxStn( g_szInTempCalCmd, INTEMPCAL_CMD_LEN ) )
        return FALSE;
    if( !GetAck() )

        return FALSE;
    if( !ReadWxStnBytes( Resp, 2 ) )
    {
        g_log.PostMsg( HOSTMSG_READWXSTNBYTES_FAIL, NULL );
        return FALSE;
    }
    g_nCalTempIn = (short int) (Resp[0] | (Resp[1] << 8));
    g_Sensors.bInsideTemp = TRUE;

    //// Outside Temperature

    if( !WriteWxStn( g_szOutTempCalCmd, OUTTEMPCAL_CMD_LEN ) )
        return FALSE;
    if( !GetAck() )
        return FALSE;
    if( !ReadWxStnBytes( Resp, 2 ) )
    {
        g_log.PostMsg( HOSTMSG_READWXSTNBYTES_FAIL, NULL );
        return FALSE;
    }
    g_nCalTempOut = (short int) (Resp[0] | (Resp[1] << 8));

    //// Barometer
    if( !WriteWxStn( g_szBaromCalCmd, BAROMCAL_CMD_LEN ) )
        return FALSE;
    if( !GetAck() )
        return FALSE;
    if( !ReadWxStnBytes( Resp, 2 ) )
    {
        g_log.PostMsg( HOSTMSG_READWXSTNBYTES_FAIL, NULL );
        return FALSE;
    }
    g_nCalBarom = (short int) (Resp[0] | (Resp[1] << 8));
    g_Sensors.bBarom = TRUE;

    //// Outside Humidity
    if( g_Sensors.bOutsideHumid )
    {
        if( !WriteWxStn( g_szOutHumCalCmd, OUTHUMCAL_CMD_LEN ) )
            return FALSE;
        if( !GetAck() )
            return FALSE;
        if( !ReadWxStnBytes( Resp, 2 ) )
        {
            g_log.PostMsg( HOSTMSG_READWXSTNBYTES_FAIL, NULL );
            return FALSE;
        }
        g_nCalHumidOut = (short int) (Resp[0] | (Resp[1] << 8));
    }

    //// get prelim readings for rate initialization, where required

    nAttempts = 3;
    while( !GetBlock(DataBlock) )
    {
        nAttempts--;
        if( nAttempts == 0 )
        {
            g_log.PostMsg( HOSTMSG_GETBLOCK_FAIL, NULL );
            return FALSE;
        }
        ResetLink();
    }

    nVal = (short int) (DataBlock[TOUT] | (DataBlock[TOUT+1] << 8));
    nVal = FixValue( nVal + g_nCalTempOut );
    TempInitSet( nVal );

    nVal = (short int) (DataBlock[BAR] | (DataBlock[BAR+1] << 8));
    nVal = FixValue( nVal - g_nCalBarom );
    nVal = MillibarsT( nVal );

    BaromInitSet( nVal );

    nVal = (short int) DataBlock[HOUT];
    nVal += g_nCalHumidOut;
    if( nVal <= 100 )
        HumidInitSet( nVal );

    // worker thread

    if( !Dav2StartThread( bVerbose ) )
        return FALSE;

    return TRUE;

}


////////////////////////////////////////////////////////////////////
// GetAck - get command ACK.  Returns TRUE success, FALSE failure
//

BOOL GetAck( void )
{
    unsigned char Byte;

    if( ReadWxStnBytes( &Byte, 1 ) )
    {
        if( Byte == ACK )
            return TRUE;
    }

    g_log.PostMsg( HOSTMSG_READWXSTNBYTES_FAIL, NULL );

    return FALSE;

}


//////////////////////////////////////////////////////////////////////////////
// ResetLink - used to insure link is in correct mode to receive commands and
// send data.  Flushes buffer in process.  If last character is NACK, returns
// TRUE (last char should be NACK), otherwise FALSE.
//

BOOL ResetLink( void )

{

    unsigned char Byte;
    char szCR[] = "\r";

    // send "RESET"
    if( !WriteWxStn( szCR, 1 ) )
        return FALSE;

    // loop until no more
    while( ReadWxStnBytes( &Byte, 1 ) );

    // check last char received
    if( Byte == NACK )
        return TRUE;
    else
        return FALSE;

}


//////////////////////////////////////////////////////////////////////////////
// start/stop worker thread that collects line of data from wx station
//

BOOL Dav2StartThread( BOOL bVerbose )
{
    int nStatus;

    if( bVerbose )
        fprintf( stdout, "Starting worker thread..." );

    nStatus = pthread_create( &g_hThread, NULL, &Dav2Thread, NULL );

    if( nStatus == 0 ) {
        if( bVerbose )
            fprintf( stdout, "done.\n" );
        return TRUE;
        }

    if( bVerbose )
        fprintf( stderr, "\n" );

    fprintf( stderr, "Failed to start worker thread.\n" );
    return FALSE;

}


void Dav2StopThread( BOOL bVerbose )
{

    g_bStopThread = TRUE;

    if( bVerbose )
        fprintf( stdout, "attempting to stop worker thread..." );

    pthread_join( g_hThread, NULL );

    FlushWxStnBuffer();

    if( bVerbose )
        fprintf( stdout, "done.\n" );

    g_bStopThread = FALSE;

}


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// Dav2Thread - thread that retrieves auto send data from the Davis station.
//

#define WAIT_TIME 3  // number of seconds between data polls

void * Dav2Thread( void * pParam )
{

    byte_t DataBlock[BLOCK_SIZE];
    int nSpeed, nDir;
    int nTemp;
    int i;
    timespec wait;
    void * pBlackHole;

    pBlackHole = pParam;

    wait.tv_sec = 1;
    wait.tv_nsec = 0;

    //FOREVER
    for(;;)
    {

        pthread_mutex_lock( &mx_Dav2Lock );

        ////////////////////////// get data packet from station

        if( GetBlock(DataBlock) )
        {

            ////////////// Wind

            nSpeed = (short int) DataBlock[WSPD];
            if( nSpeed )
            {
                g_nWindCount++;

                // accumulate for average
                g_nSpeedAccum += nSpeed;

                nDir = (short int) (DataBlock[WDIR] | (DataBlock[WDIR+1] << 8));
                g_nDirAccum += nDir;
                // need to track quadrants to correct for QI/QIV ambiguity
                if( nDir < 90 )
                    g_nDav2Q1++;
                else if ( nDir > 270 )
                    g_nDav2Q4++;

                if( nSpeed >= g_nGustSpeed )
                {
                    g_nGustSpeed = nSpeed;
                    g_nGustDir = nDir;
                }
            }

            ////////////// Temperature

            g_bTempRead = TRUE;

            nTemp = (short int) (DataBlock[TOUT] | (DataBlock[TOUT+1] << 8));
            nTemp = FixValue( nTemp + g_nCalTempOut );

            if( nTemp >= g_nHiTemp )
                g_nHiTemp = nTemp;

            if( nTemp <= g_nLoTemp )
                g_nLoTemp = nTemp;

        }
        else
            ResetLink();

        pthread_mutex_unlock( &mx_Dav2Lock );

        ////////////////////////// wait/check

        for( i=0; i<WAIT_TIME; i++ )
        {
            if( g_bStopThread )
                return NULL;

            nanosleep( &wait, NULL );
        }

    }

    return NULL;

}

