/***************************************************************************
 *   Copyright (C) 2005 by John Bennett   *
 *   n4xi@arrl.net   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

//
// mresolv.cpp  -  email address resolver routines
//

#include <stdlib.h>
#include <stdio.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include <memory.h>
#include <ctype.h>

#include "mresolv.h"

#ifndef TRUE
    #define TRUE    1
    #define FALSE   0
#endif

#ifndef BOOL
    typedef int BOOL;
#endif

#ifndef FOREVER
    #define FOREVER for(;;)
#endif


#define ANS_SIZE 512

#define ID_OFFSET       0
#define FLAGS_OFFSET    2
#define QDCOUNT_OFFSET  4
#define ANCOUNT_OFFSET  6
#define NSCOUNT_OFFSET  8
#define ARCOUNT_OFFSET  10

struct msgheader
{
    unsigned short uID;
    unsigned short uFlags;
    unsigned short uQdCount;     // question count (usually one?)
    unsigned short uAnCount;     // answer count
    unsigned short uNsCount;     // number of name server resource records in authority records
    unsigned short uArCount;     // number of resource records in additional records section
};
#define MSGHEADER_SIZE sizeof( struct msgheader )

struct msgheader_flags
{
    int nQR;        // Query/Response
    int nOpcode;
    BOOL bAA;       // authoritative answer
    BOOL bTC;       // truncated
    BOOL bRD;       // recursion desired
    BOOL bRA;       // recursion available
    int nRespCode;  // response code
};
#define MSGHEADER_FLAGS_SIZE sizeof( struct msgheader_flags )

// opcodes
#define O_QUERY   0   // standard query
#define O_IQUERY  1   // inverse query
#define O_STATUS  2

#define OPCODE_MASK 0x7800
#define OPCODE_SHIFT 11

// misc flags
#define QR_MASK 0x8000  // query/response
#define AA_MASK 0x0400  // authoritative answer
#define TC_MASK 0x0200  // truncated
#define RD_MASK 0x0100  // recursion desired
#define RA_MASK 0x0080  // recursion available
#define Z_MASK 0x0070   // must be zero (ignore ?)

// response codes
#define R_NO_ERR    0
#define R_FMT_ERR   1
#define R_SRV_FAIL  2
#define R_NAME_ERR  3
#define R_NOT_IMPL  4   // not implemented
#define R_REFUSED   5

#define RCODE_MASK  0x000f  // response code

//qtypes
#define A       1   // host address
#define NS      2   // name server
#define CNAME   5   // canonical name for alias
#define SOA     6   // start of authority for zone
#define WKS     11  // well-known service
#define PTR     12  // domain name pointer
#define HINFO   13  // host info
#define MINFO   14  // mailbox or mail list info
#define MX      15  // mail exchange
#define TXT     16  // text strings

//qclass
#define IN  1   // internet
#define CH  3   // chaos (MIT)
#define HS  4   // Hesiod (Dyer 87)

// local prototypes
unsigned char * ExtractDomain( char * pDomain, unsigned char * pS1, unsigned char * pAns );


BOOL ResolveMXDomain( mailform * pMail )
{
    msgheader Header;
    msgheader_flags Flags;
    unsigned char * pS;
    unsigned char ans[ANS_SIZE];
    int nType;
    int nPref;
    int nMxPref;
    unsigned short i;
    in_addr_t inaddr;
    char AnRrData[128];
    char sTemp[128];

    memset( &pMail->inaddr, 0, 4 );

    if( res_search( pMail->sToHost, C_IN, T_MX, ans, ANS_SIZE ) == -1 )
        return FALSE;

    // big endian order
    Header.uID = (unsigned short) ((ans[0]<<8) + ans[1]);
    Header.uFlags = (unsigned short) ((ans[FLAGS_OFFSET]<<8) + ans[FLAGS_OFFSET+1]);
    Header.uQdCount = (int) ((ans[QDCOUNT_OFFSET]<<8) + ans[QDCOUNT_OFFSET+1]);
    Header.uAnCount = (int) ((ans[ANCOUNT_OFFSET]<<8) + ans[ANCOUNT_OFFSET+1]);
    Header.uNsCount = (int) ((ans[NSCOUNT_OFFSET]<<8) + ans[NSCOUNT_OFFSET+1]);
    Header.uArCount = (int) ((ans[ARCOUNT_OFFSET]<<8) + ans[ARCOUNT_OFFSET+1]);

    if( Header.uAnCount == 0 )
        return FALSE;
    if( Header.uArCount == 0 )
        return FALSE;
    if( !Header.uFlags & QR_MASK )
        return FALSE;

    Flags.nOpcode = (Header.uFlags & OPCODE_MASK) >> OPCODE_SHIFT;

    if( Header.uFlags & AA_MASK )       // authoritative answer
        Flags.bAA = TRUE;
    else
        Flags.bAA = FALSE;

    if( !Header.uFlags & TC_MASK )      // answer truncated
        return FALSE;
    if( Header.uFlags & RCODE_MASK )    // response code must be zero (OK)
        return FALSE;

    pS = &ans[MSGHEADER_SIZE];
    pS = ExtractDomain( NULL, pS, ans );
    pS += 4;    // Qtype, QClass

    //////////////////////// answer
    nMxPref = 99999;  // be certain first found takes
    for(i=0; i< Header.uAnCount; i++ )
    {
        pS = ExtractDomain( NULL, pS, ans );
        nType = (int) ((pS[0]<<8) | pS[1]);
        if( nType != MX )
            return FALSE;
        pS += 10; // rrtype, rrclass, rrttl, rrdlen
        nPref = (int) ((pS[0]<<8) + pS[1]); // mxpref
        pS += 2;
        if( nPref < nMxPref )
        {
            nMxPref = nPref;
            pS = ExtractDomain( AnRrData, pS, ans );
        }
        else
            pS = ExtractDomain( NULL, pS, ans );
    }

    //////////////////////// name servers (totally useless, but must do anyway)
    for( i=0; i< Header.uNsCount; i++ )
    {
        pS = ExtractDomain( NULL, pS, ans );
        pS += 10;   // rrtype, rrdlen, rrclass, rrttl
        pS = ExtractDomain( NULL, pS, ans );
    }

    //////////////////////// additional records section
    inaddr = (in_addr_t) 0;
    for( i=0; i< Header.uArCount; i++ )
    {
        pS = ExtractDomain( sTemp, pS, ans );
        nType = (int) ((pS[0]<<8) | pS[1]);
        if( nType != A )
            return FALSE;   // something wrong
        pS += 10; // rrtype, rrclass, rrttl, rrdlen
        if( strcmp( sTemp, AnRrData ) == 0 )
        {
            memcpy( &inaddr, pS, 4 );
            break;  // no point continuing since we now have the address we need
        }
        pS += 4; // rrdata
    }

    if( inaddr )
    {
        memcpy( &pMail->inaddr, &inaddr, 4 );
        return TRUE;
    }

    return FALSE;
}


//////////////////////////////////////////////////////////////////////////////
// ExtractDomain - recursively builds domain name, including compressed
// names. Returns location of next byte after domain definition. If pDomain
// is NULL, does not copy - just advances the pointer to the message stream.
//

unsigned char * ExtractDomain( char * pDomain, unsigned char * pS1, unsigned char * pAns )
{
    char * pD;
    unsigned short uOffset;
    int nLen;

    pD = pDomain;

    FOREVER
    {
        if( (pS1[0] & 0xc0) == 0xc0 )
        {
            uOffset = (unsigned short)((pS1[0]<<8) + pS1[1]) & 0x3fff;
            ExtractDomain( pD, (pAns + uOffset), pAns );
            return pS1 + 2;
        }
        else if ( pS1[0] )
        {
            nLen = (unsigned short) pS1[0];
            pS1++;
            if( pD )
            {
                memcpy( pD, pS1, nLen );
                pD += nLen;
                *pD++ = '.';
            }
            pS1 += nLen;
        }
        else
        {
            pS1++;  // point past zero terminator
            break;
        }
    }

    if( pD )
        *--pD = 0;

    return pS1;
}


//////////////////////////////////////////////////////////////////////////////
// IsIP - returns TRUE if string is an IP address in dotted decimal format,
// FALSE otherwise.
//

BOOL IsIP( char * pAddr )
{
    char * pS;

    pS = pAddr;

    while( *pS )
    {
        if( !isdigit( (int)*pS ) )
            return FALSE;
        pS++;
        if( *pS == '.' )
            return TRUE;
    };

    return FALSE;
}

