/***************************************************************************
 *   Copyright (C) 2005 by John Bennett   *
 *   n4xi@arrl.net   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

//
// onewire.cpp - base one-wire routines
//

#define _ONEWIRE_CPP_

#include <string.h>
#include <math.h>

// libwxn
#include <logmsg.h>

#include "hconfig.h"
#include "wxdata.h"
#include "ownet.h"
#include "findtype.h"
#include "cnt1d.h"
#include "temp10.h"
#include "atod26.h"
#include "swt05.h"
#include "onewire.h"


#define OW_DS2423_COUNTER  0x1d
#define OW_DS2405_PIO      0x05
#define OW_DS1820_TEMP     0x10
#define OW_DS2438_AD       0x26

#define OW_SN_LEN           8
#define MAXDEVICES          10

static BOOL g_bOneWireOpened = FALSE;

static uchar g_EmpRadSN[OW_SN_LEN];
static uchar g_AspCtrlSN[OW_SN_LEN];
static uchar g_HtrCtrlSN[OW_SN_LEN];
static uchar g_SoilTempSN[OW_SN_LEN];
static uchar g_Soil2TempSN[OW_SN_LEN];
static uchar g_InsideTempSN[OW_SN_LEN];
static uchar g_OutsideTempSN[OW_SN_LEN];
static uchar g_InsideHumidSN[OW_SN_LEN];
static uchar g_OutsideHumidSN[OW_SN_LEN];
static uchar g_BarometerSN[OW_SN_LEN];
//static uchar g_SunDurationSN[OW_SN_LEN];
//static uchar g_UvSN[OW_SN_LEN];


////////////////////////////////////////////////////////////////////
// OpenOneWire - open the serial port for the one-wire devices
//

BOOL OpenOneWire( char * szPort )
{
    char szDev[32];

    sprintf( szDev, "/dev/%s", szPort );

    // attempt to acquire the 1-Wire Net
    g_bOneWireOpened = owAcquire( szDev );

    if( g_bOneWireOpened )
        return TRUE;
    else
        return FALSE;
}


////////////////////////////////////////////////////////////////////
// CloseOneWire - close the serial port for the one-wire devices
//

void CloseOneWire( void )
{
    if( g_bOneWireOpened )
    {
        // turn off controls
        OwCtrlReset();

        g_bOneWireOpened = 0;
        owRelease();
    }
}



////////////////////////////////////////////////////////////////////
// IdOneWireDevices - ID all one-wire devices. If not found,
// set config parms to FALSE for sensors and appropriate message to
// screen and log. If at least one flag is TRUE, routine returns
// TRUE, otherwise FALSE.
//

BOOL IdOneWireDevices( void )
{
    int nDevices;
    uchar FamilySN[MAXDEVICES][OW_SN_LEN];
    BOOL bFound;
    BOOL bSNRequired;

    bFound = FALSE;

    // EMP/RM-60 interface
    if( g_HostConfig.bOwEmp || g_HostConfig.bOwRad )
    {
        nDevices = FindDevices( &FamilySN[0], OW_DS2423_COUNTER, MAXDEVICES );
        if ( nDevices == 1 )
        {
            memcpy( &g_EmpRadSN, &FamilySN[0], OW_SN_LEN );
            bFound = TRUE;
        }
        else
        {  // none or more than one
            g_log.PostMsg( HOSTMSG_DS2423_FAIL, "OW EMP/Radiation sensor disabled" );
            g_HostConfig.bOwEmp = FALSE;
            g_HostConfig.bOwRad = FALSE;
        }
    }

    // temperature
    nDevices = FindDevices( &FamilySN[0], OW_DS1820_TEMP, MAXDEVICES );
    if( nDevices )
    {
        if( nDevices > 1 )
            bSNRequired = TRUE;
        else
            bSNRequired = FALSE;

        // soil temp sensor
        if( g_HostConfig.bOwSoilTemp )
        {
            if( bSNRequired )
            {
                if( *g_HostConfig.OwSoilTempSN )
                {
                    memcpy( &g_SoilTempSN, &g_HostConfig.OwSoilTempSN, OW_SN_LEN );
                    bFound = TRUE;
                }
                else
                {
                    g_log.PostMsg( HOSTMSG_DS18S20_SN_REQ, "OW Soil temp sensor disabled" );
                    g_HostConfig.bOwSoilTemp = FALSE;
                }
            }
            else if ( nDevices == 1 )
            {
                memcpy( &g_SoilTempSN, &FamilySN[0], OW_SN_LEN );
                bFound = TRUE;
                nDevices = 0;   // be certain next sensor does not use the same device
            }
            else
            {  // none or more than one
                g_log.PostMsg( HOSTMSG_DS18S20_SN_REQ, "OW Soil temp sensor disabled" );
                g_HostConfig.bOwSoilTemp = FALSE;
            }
        }

        // soil2 temp sensor
        if( g_HostConfig.bOwSoil2Temp )
        {
            if( bSNRequired )
            {
                if( *g_HostConfig.OwSoil2TempSN )
                {
                    memcpy( &g_Soil2TempSN, &g_HostConfig.OwSoil2TempSN, OW_SN_LEN );
                    bFound = TRUE;
                }
                else
                {
                    g_log.PostMsg( HOSTMSG_DS18S20_SN_REQ, "OW Soil2 temp sensor disabled" );
                    g_HostConfig.bOwSoil2Temp = FALSE;
                }
            }
            else if ( nDevices == 1 )
            {
                memcpy( &g_Soil2TempSN, &FamilySN[0], OW_SN_LEN );
                bFound = TRUE;
                nDevices = 0;   // be certain next sensor does not use the same device
            }
            else
            {  // none or more than one
                g_log.PostMsg( HOSTMSG_DS18S20_SN_REQ, "OW Soil2 temp sensor disabled" );
                g_HostConfig.bOwSoil2Temp = FALSE;
            }
        }

        // inside temp sensor
        if( g_HostConfig.bOwInsideTemp )
        {
            if( bSNRequired )
            {
                if( *g_HostConfig.OwInsideTempSN )
                {
                    memcpy( &g_InsideTempSN, &g_HostConfig.OwInsideTempSN, OW_SN_LEN );
                    bFound = TRUE;
                }
                else
                {
                    g_log.PostMsg( HOSTMSG_DS18S20_SN_REQ, "Inside temp sensor disabled" );
                    g_HostConfig.bOwInsideTemp = FALSE;
                }
            }
            else if ( nDevices == 1 )
            {
                memcpy( &g_InsideTempSN, &FamilySN[0], OW_SN_LEN );
                bFound = TRUE;
                nDevices = 0;   // be certain next sensor does not use the same device
            }
            else
            {  // none or more than one
                g_log.PostMsg( HOSTMSG_DS18S20_SN_REQ, "Inside temp sensor disabled" );
                g_HostConfig.bOwInsideTemp = FALSE;
            }
        }

        // outside temp sensor
        if( g_HostConfig.bOwOutsideTemp )
        {
            if( bSNRequired )
            {
                if( *g_HostConfig.OwOutsideTempSN )
                {
                    memcpy( &g_OutsideTempSN, &g_HostConfig.OwOutsideTempSN, OW_SN_LEN );
                    bFound = TRUE;
                }
                else
                {
                    g_log.PostMsg( HOSTMSG_DS18S20_SN_REQ, "Outside temp sensor disabled" );
                    g_HostConfig.bOwOutsideTemp = FALSE;
                }
            }
            else if ( nDevices == 1 )
            {
                memcpy( &g_OutsideTempSN, &FamilySN[0], OW_SN_LEN );
                bFound = TRUE;
                nDevices = 0;   // be certain next sensor does not use the same device
            }
            else
            {  // none or more than one
                g_log.PostMsg( HOSTMSG_DS18S20_SN_REQ, "Outside temp sensor disabled" );
                g_HostConfig.bOwOutsideTemp = FALSE;
            }
        }
    }

    // control interfaces
    nDevices = FindDevices( &FamilySN[0], OW_DS2405_PIO, MAXDEVICES );
    if( nDevices )
    {
        if( nDevices > 1 )
            bSNRequired = TRUE;
        else
            bSNRequired = FALSE;

        // aspirator control
        if( g_HostConfig.bOwAspCtrl )
        {
            if( bSNRequired )
            {
                if( *g_HostConfig.OwAspCtrlSN )
                {
                    memcpy( &g_AspCtrlSN, &g_HostConfig.OwAspCtrlSN, OW_SN_LEN );
                    bFound = TRUE;
                }
                else
                {
                    g_log.PostMsg( HOSTMSG_DS2405_SN_REQ, "OW aspirator control disabled" );
                    g_HostConfig.bOwAspCtrl = FALSE;
                }
            }
            else if ( nDevices == 1 )
            {
                memcpy( &g_AspCtrlSN, &FamilySN[0], OW_SN_LEN );
                bFound = TRUE;
                nDevices = 0;   // be certain next sensor does not use the same device
            }
            else
            {  // none or more than one
                g_log.PostMsg( HOSTMSG_DS2405_SN_REQ, "OW aspirator control disabled" );
                g_HostConfig.bOwAspCtrl = FALSE;
            }
        }

        // rain gauge heater control
        if( g_HostConfig.bOwHtrCtrl )
        {
            if( bSNRequired )
            {
                if( *g_HostConfig.OwHtrCtrlSN )
                {
                    memcpy( &g_HtrCtrlSN, &g_HostConfig.OwHtrCtrlSN, OW_SN_LEN );
                    bFound = TRUE;
                }
                else
                {
                    g_log.PostMsg( HOSTMSG_DS2405_SN_REQ, "Rain gauge heater control disabled" );
                    g_HostConfig.bOwHtrCtrl = FALSE;
                }
            }
            else if ( nDevices == 1 )
            {
                memcpy( &g_HtrCtrlSN, &FamilySN[0], OW_SN_LEN );
                bFound = TRUE;
                nDevices = 0;   // be certain next sensor does not use the same device
            }
            else
            {  // none or more than one
                g_log.PostMsg( HOSTMSG_DS2405_SN_REQ, "Rain gauge heater control disabled" );
                g_HostConfig.bOwHtrCtrl = FALSE;
            }

        }

    }

    // barometer, humidity sensors
    nDevices = FindDevices( &FamilySN[0], OW_DS2438_AD, MAXDEVICES );
    if( nDevices )
    {
        if( nDevices > 1 )
            bSNRequired = TRUE;
        else
            bSNRequired = FALSE;

        // barometric pressure sensor
        if( g_HostConfig.bOwBarom )
        {
            if( bSNRequired )
            {
                if( *g_HostConfig.OwBaromSN )
                {
                    memcpy( &g_BarometerSN, &g_HostConfig.OwBaromSN, OW_SN_LEN );
                    bFound = TRUE;
                }
                else
                {
                    g_log.PostMsg( HOSTMSG_DS2438_SN_REQ, "OW Barometer disabled" );
                    g_HostConfig.bOwBarom = FALSE;
                }
            }
            else if ( nDevices == 1 )
            {
                memcpy( &g_BarometerSN, &FamilySN[0], OW_SN_LEN );
                bFound = TRUE;
                nDevices = 0;   // be certain next sensor does not use the same device
            }
            else
            {  // none or more than one
                g_log.PostMsg( HOSTMSG_DS2438_SN_REQ, "OW Barometer disabled" );
                g_HostConfig.bOwBarom = FALSE;
            }
        }

        // outside humidity
        if( g_HostConfig.bOwOutsideHumid )
        {
            if( bSNRequired )
            {
                if( *g_HostConfig.OwOutsideHumidSN )
                {
                    memcpy( &g_OutsideHumidSN, &g_HostConfig.OwOutsideHumidSN, OW_SN_LEN );
                    bFound = TRUE;
                }
                else
                {
                    g_log.PostMsg( HOSTMSG_DS2438_SN_REQ, "OW outside humidity sensor disabled" );
                    g_HostConfig.bOwOutsideHumid = FALSE;
                }
            }
            else if ( nDevices == 1 )
            {
                memcpy( &g_OutsideHumidSN, &FamilySN[0], OW_SN_LEN );
                bFound = TRUE;
                nDevices = 0;   // be certain next sensor does not use the same device
            }
            else
            {  // none or more than one
                g_log.PostMsg( HOSTMSG_DS2438_SN_REQ, "OW outside humidity sensor disabled" );
                g_HostConfig.bOwOutsideHumid = FALSE;
            }
        }

        // inside humidity
        if( g_HostConfig.bOwInsideHumid )
        {
            if( bSNRequired )
            {
                if( *g_HostConfig.OwInsideHumidSN )
                {
                    memcpy( &g_InsideHumidSN, &g_HostConfig.OwInsideHumidSN, OW_SN_LEN );
                    bFound = TRUE;
                }
                else
                {
                    g_log.PostMsg( HOSTMSG_DS2438_SN_REQ, "OW inside humidity sensor disabled" );
                    g_HostConfig.bOwInsideHumid = FALSE;
                }
            }
            else if ( nDevices == 1 )
            {
                memcpy( &g_InsideHumidSN, &FamilySN[0], OW_SN_LEN );
                bFound = TRUE;
                nDevices = 0;   // be certain next sensor does not use the same device
            }
            else
            {  // none or more than one
                g_log.PostMsg( HOSTMSG_DS2438_SN_REQ, "OW inside humidity sensor disabled" );
                g_HostConfig.bOwInsideHumid = FALSE;
            }
        }
    }

    return bFound;
}


//////////////////////////////////////////////////////////////////////////////
// OwReadCounter - returns count from one of the dual counters specified.
//

BOOL OwReadCounter( int nType, unsigned long & ulCount )
{
    uchar * pSN;
    int nPage;

    switch( nType )
    {
        case EMP_CNT:
            pSN = g_EmpRadSN;
            nPage = 15;
            break;

        case RAD_CNT:
            pSN = g_EmpRadSN;
            nPage = 14;
            break;

        default:
            return 0;
    }

    if( ReadCounter(pSN, nPage, &ulCount) )
        return TRUE;
    else
    {
        g_log.PostMsg( HOSTMSG_OWREADCOUNTER_FAIL, NULL );
        return FALSE;
    }
}


//////////////////////////////////////////////////////////////////////////////
// OwReadTemp - returns temperature read from specified DS18S20.
//

BOOL OwReadTemp( int nType, int & nVal )
{
    int nTemp;
    uchar * pSN;

    switch( nType )
    {
        case INSIDE_TEMP:
            pSN = g_InsideTempSN;
            break;

        case SOIL_TEMP:
            pSN = g_SoilTempSN;
            break;

        case SOIL2_TEMP:
            pSN = g_SoilTempSN;
            break;

        case OUTSIDE_TEMP:
            pSN = g_OutsideTempSN;
            break;

        default:
            return FALSE;

    }

    if( DS1820Temperature(pSN, &nTemp) )
    {
        nVal = nTemp;
        return TRUE;
    }

    g_log.PostMsg( HOSTMSG_OWREADTEMP_FAIL, NULL );
    return FALSE;
}


//////////////////////////////////////////////////////////////////////////////
// OwReadBarom - reads raw A/D voltage from OW, converts to tenths mb and
// returns converted value.
//

int OwReadBarom( void )
{
    int nBarom;
    int nVal;

    nVal = ReadAtoD( FALSE, &g_BarometerSN[0] );

    if( nVal == 0 )
        g_log.PostMsg( HOSTMSG_OWREADBAROM_FAIL, NULL );

    nBarom = ((nVal - 175)/4 + 870)*10;

    switch( (nVal - 175) % 4 )
    {
        case 0:
            break;

        case 1:
            nBarom += 3;    // 0.3 (0.25)
            break;

        case 2:
            nBarom += 5;    // 0.5
            break;

        case 3:
            nBarom += 8;    // 0.8 (0.75)
            break;
    }

    nBarom += g_HostConfig.nBaromOffset;

    return nBarom;
}


//////////////////////////////////////////////////////////////////////////////
// OwReadHumid - reads specified HIH-3610 on OW bus and returns % humidity
// returns TRUE success, FALSE failure.
//

BOOL OwReadHumid( int nType, int & nVal )
{
    int nHumid;
    int nVdd;
    int nVad;
    int nTemp;
    double dTempComp;
    double dRes;
    uchar * pSN;
    int i;

    if( nType == INSIDE_HUMIDITY )
        pSN = g_InsideHumidSN;
    else    // OUTSIDE_HUMIDITY:
        pSN = g_OutsideHumidSN;

    // supply voltage in 10's of mV
    nVdd = ReadAtoD( TRUE, pSN );
    if( (nVdd > 580) || (nVdd < 400) )
    {
        g_log.PostMsg( HOSTMSG_OWREADHUMID_FAIL, "Vdd outside accepted limits" );
        return FALSE;
    }

    // sensor voltage in 10's of mV
    nVad = ReadAtoD( FALSE, pSN );

    i=0;
    //FOREVER
    for(;;)
    {
        if( DS2438Temperature( pSN, &nTemp ) )
        {
            dTempComp = 1.0546 - 0.00216 * (double) nTemp;
            break;
        }
        if( ++i == 3 )  // max tries exceeded
        {
            g_log.PostMsg( HOSTMSG_OWREADHUMID_FAIL, "DS2438Temperature(), no compensation will be used" );
            dTempComp = 1;
            break;
        }
    }

    dRes = ( ( ( (double)nVad/(double)nVdd ) - 0.16)/0.0062 )/dTempComp;
    if( dRes > 100 )
    {
        nVal = 100;
        return TRUE;
    }
    else if( dRes < 0 )
    {
        g_log.PostMsg( HOSTMSG_OWREADHUMID_FAIL, "result < 0" );
        return FALSE;
    }

    nHumid = (int) dRes;
    if( (dRes + 0.5) >= (double) (nHumid + 1) )
        nHumid++;

    nVal = nHumid;

    return TRUE;
}


//////////////////////////////////////////////////////////////////////////////
// OwCtrlStatus - read state of control output on OW interface board. Returns
// on (1), off (0) or -1 on error.
//

int OwCtrlStatus( int nType )
{
    int nLevel;
    uchar * pSN;

    if( nType == ASPIRATOR )
        pSN = g_AspCtrlSN;
    else    // HEATER
        pSN = g_HtrCtrlSN;

    if( ReadSwitch05( pSN, &nLevel ) )
        return nLevel;

    else
        return -1;
}


//////////////////////////////////////////////////////////////////////////////
// OwCtrlSet - Sets control on (0) or off (1). Returns TRUE success,
// FALSE failure.
//

BOOL OwCtrlSet( int nType, int nLevel )
{
    uchar * pSN;

    if( nType == ASPIRATOR )
        pSN = g_AspCtrlSN;
    else    // HEATER
        pSN = g_HtrCtrlSN;

    return SetSwitch05( pSN, nLevel );
}


//////////////////////////////////////////////////////////////////////////////
// OwCtrlReset - if configured and detected, turn off control ports
//

void OwCtrlReset( void )
{
    if( g_HostConfig.bOwAspCtrl )
        OwCtrlSet( ASPIRATOR, OFF );

    if( g_HostConfig.bOwHtrCtrl )
        OwCtrlSet( HEATER, OFF );
}


//////////////////////////////////////////////////////////////////////////////
// OwControl - Sets control on (1) or off (0) for aspirator and/or rain
// gauge heater.
//

void OwControl( void )
{
    if( g_HostConfig.bOwAspCtrl )
    {
        if( g_Local.Data.hcr > g_HostConfig.nAspiratorTrip )
            OwCtrlSet( ASPIRATOR, ON );
        else
            OwCtrlSet( ASPIRATOR, OFF );
    }

    if( g_HostConfig.bOwHtrCtrl )
    {
        if( g_Local.Data.tcr < 35 )
            OwCtrlSet( HEATER, ON );
        else
            OwCtrlSet( HEATER, OFF );
    }
}

