/***************************************************************************
 *   Copyright (C) 2005 by John Bennett   *
 *   n4xi@arrl.net   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#define _PRECIP_CPP_

#include <stdlib.h>
#include <libpq-fe.h>

// libwxn
#include <tz.h>

#include "wxdata.h"
#include "rates.h"
#include "precip.h"

// local prototypes
int PrecipDiff( int nCurrent );

// local variables
static int g_nLastPrecip = 0;


//////////////////////////////////////////////////////////////////////////////
// SetLastPrecip
//

void SetLastPrecip( int nVal )

{
    g_nLastPrecip = nVal;
}

//////////////////////////////////////////////////////////////////////////////
// SetPrecipLastHour - sets <plh> variable in database to total precip
// that has occurred in the last 60 minutes as recorded on the hour.
//

void SetPrecipLastHour( void )
{

    PGconn * pConn;
    PGresult * pResult;
    char szExec[128];

    pConn = PQconnectdb( "dbname=wxn" );
    if( pConn )
    {
        if( PQstatus( pConn ) == CONNECTION_OK )
        {
            sprintf( szExec, "SELECT sum(plt) FROM precip WHERE hour=%i", (g_Time.Hour()+23)%24 );
            pResult = PQexec( pConn, szExec );
            if( PQresultStatus( pResult ) == PGRES_TUPLES_OK )
            {
                if( PQntuples( pResult ) )
                    g_Local.Data.plh = atoi( PQgetvalue( pResult, 0, 0 ) );
                else
                    g_Local.Data.plh = 0;
            }

            PQclear( pResult );
        }

        PQfinish( pConn );

    }



}


//////////////////////////////////////////////////////////////////////////////
// PrecipSet - sets all precip related values. Argument supplied is current
// gauge reading. Correction, if required, is handled by PrecipDiff(), called
// within this routine. This function is called in the driver-specific
// implementation file by all drivers. It is expected that the value handed
// to this routine is the current gauge reading in .01" increments (as an
// integer).
//

void PrecipSet( int nPrecip )
{

    int nMin;
    int nDiff;
    PGconn * pConn;
    PGresult * pResult;
    char szExec[128];


    // get reading difference and fix underflow if it occurs
    nDiff = PrecipDiff( nPrecip );

    // daily total
    g_Local.Data.pca += nDiff;

    // update monthly, annual totals
    g_Local.Data.pta += nDiff;
    g_Local.Data.ptm += nDiff;

    // instantaneous rain rate
    g_Local.Data.pri = nDiff * HOUR_INTERVAL_COUNT;

    // hourly rate
    pConn = PQconnectdb( "dbname=wxn" );
    if( pConn )
    {
        if( PQstatus( pConn ) == CONNECTION_OK )
        {
            // force minute to be multiple of 5
            nMin = (g_Time.Minute()/5) * 5;

            // save current reading
            sprintf( szExec, "UPDATE precip SET plt=%i WHERE hour=%i AND min=%i",
                nDiff, g_Time.Hour(), nMin );
            pResult = PQexec( pConn, szExec );
            PQclear( pResult );

            // hourly rate
            sprintf( szExec, "SELECT sum(plt) FROM precip WHERE (hour=%i AND min>%i) OR (hour=%i AND min<=%i)",
                (g_Time.Hour()+23)%24, nMin, g_Time.Hour(), nMin );
            pResult = PQexec( pConn, szExec );
            if( PQresultStatus( pResult ) == PGRES_TUPLES_OK )
            {
                if( PQntuples( pResult ) )
                    g_Local.Data.prh = atoi( PQgetvalue( pResult, 0, 0 ) );
                else
                    g_Local.Data.prh = 0;
            }
            PQclear( pResult );

            // 24 hour rate
            sprintf( szExec, "SELECT sum(plt) FROM precip" );
            pResult = PQexec( pConn, szExec );
            {
                if( PQntuples( pResult ) )
                    g_Local.Data.p24 = atoi( PQgetvalue( pResult, 0, 0 ) );
                else
                    g_Local.Data.p24 = 0;
            }
            PQclear( pResult );

        }

        PQfinish( pConn );

    }

}


//////////////////////////////////////////////////////////////////////////////
// PrecipDiff - returns difference between current and last reading. If
// reading is reset, current accum becomes current reading. In all cases,
// g_nLastPrecip is set to nCurrent.
//

int PrecipDiff( int nCurrent )

{

    int nDiff;

    if( nCurrent < g_nLastPrecip )
        nDiff = nCurrent;
    else
        nDiff = nCurrent - g_nLastPrecip;

    g_nLastPrecip = nCurrent;

    return nDiff;

}


//////////////////////////////////////////////////////////////////////////////
// ResetMonthPrecip
//

void ResetMonthPrecip( void )

{

  g_Local.Data.ptm = 0;

}


//////////////////////////////////////////////////////////////////////////////
// ResetAnnualPrecip
//

void ResetAnnualPrecip( void )

{

  g_Local.Data.pta = 0;

}

