/***************************************************************************
 *   Copyright (C) 2005 by John Bennett   *
 *   n4xi@arrl.net   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

//
// wxdata.cpp - global weather data used by wxnhostd and related routines.
//

#define _WXDATA_CPP_

#include <stdlib.h>
#include <memory.h>
#include <libpq-fe.h>

// libwxn
#include <wxcalc.h>
#include <logmsg.h>
#include <sun.h>
#include <tz.h>

#include "wxdata.h"
#include "umx.h"
#include "id5.h"
#include "dav2.h"
#include "davp.h"
#include "hglobal.h"
#include "hconfig.h"
#include "sensors.h"
#include "emp.h"
#include "precip.h"
#include "barom.h"
#include "humid.h"
#include "temp.h"
#include "rad.h"
#include "onewire.h"
#include "rates.h"


struct sensors g_Sensors;
struct nodeblk g_Local;


//////////////////////////////////////////////////////////////////////////////
// GetCurrentWeatherData - called with bNormal==TRUE for normal reads; FALSE
// when called for initialization.
//

BOOL GetCurrentWeatherData( BOOL bNormal )
{
    BOOL bStatus;
    int nVal;

    switch( g_HostConfig.nStnType )
    {
        case ULT2000:
        case ULTII:
            bStatus = UmxProcessData();
            break;

        case ID5001:
            bStatus = Id5ProcessData();
            break;

        case DAVIS2:
            bStatus = Dav2ProcessData();
            break;

        case DAVISVP:
            bStatus = DavPProcessData();
            break;

        default:
            return FALSE;
    }

    if( bStatus )
        CalculateFields( &g_Local.Data );
    else
        g_log.PostMsg( HOSTMSG_WXSTN_DATA_NOT_AVAILABLE, NULL );

    if( g_Sensors.bOwRad )
    {
        RadRead();
        if( bNormal )
        {
            if ( g_Local.Data.rca )
                RadCheck();
            RadAvg();
        }
    }

    if( g_Sensors.bOwEmp )
        EmpRead();

    if( g_Sensors.bOwInsideTemp )
    {
        if( OwReadTemp(INSIDE_TEMP, nVal) )
            g_Local.Data.ist = nVal;
    }

    if( g_Sensors.bOwOutsideTemp )
    {
        if( OwReadTemp(OUTSIDE_TEMP, nVal) )
            TempSet( nVal, nVal, nVal );
    }

    if( g_Sensors.bOwSoilTemp )
    {
        if( OwReadTemp(SOIL_TEMP, nVal) )
            g_Local.Data.tsd = nVal;
    }

    if( g_Sensors.bOwSoil2Temp )
    {
        if( OwReadTemp(SOIL2_TEMP, nVal) )
            g_Local.Data.tsd2 = nVal;
    }

    if( g_Sensors.bOwInsideHumid )
    {
        if( OwReadHumid( INSIDE_HUMIDITY, nVal ) )
            g_Local.Data.ish = nVal;
        else
            g_log.PostMsg( HOSTMSG_OW_IN_HUMID_READ_FAIL, NULL );
    }

    if( g_HostConfig.bOwAspCtrl || g_HostConfig.bOwHtrCtrl )
        OwControl();

    if( g_Sensors.bOwBarom )
    {
        nVal = OwReadBarom();
        if( nVal )
            BaromSet( nVal );
    }

    if( g_Sensors.bOwOutsideHumid )
    {
        if( OwReadHumid( OUTSIDE_HUMIDITY, nVal ) )
            HumidSet( nVal );
        else
            g_log.PostMsg( HOSTMSG_OW_OUT_HUMID_READ_FAIL, NULL );
    }

    return bStatus;
}


//////////////////////////////////////////////////////////////////////////////
// InitWxData - initialize database and all variables. Returns TRUE success,
// FALSE on fail.
//

BOOL InitWxData( BOOL bVerbose )
{
    int nPca;

    nPca = 0;

    // be sure we start with clean slate
    bzero( &g_Local.Data, NODEDATA_SIZE );
    bzero( &g_Sensors, SENSORS_SIZE );

    //////////////////////// init wx station

    switch( g_HostConfig.nStnType )
    {
        case ULT2000:
        case ULTII:
            if( !UmxInit( bVerbose ) )
                return FALSE;
            break;

        case ID5001:
            if( !Id5Init( bVerbose ) )
                return FALSE;
            break;

        case DAVIS2:
            if( !Dav2Init( bVerbose ) )
                return FALSE;
            break;

        case DAVISVP:
            if( !DavPInit( bVerbose ) )
                return FALSE;
            break;

        default:
            g_log.PostMsg( HOSTMSG_WXSTN_NOT_DEFINED, NULL );
            return FALSE;

    }

    if( bVerbose )
        fprintf( stdout, "Wx station driver initialization completed.\n" );

    //////////////////////// data from backend

    // assume there is no previous data

    if( g_HostConfig.bEnableOneWire )
    {
        SetOwSensors( bVerbose );
        OwSensorsInit();
    }
    if( g_HostConfig.bNullNoSensors )
        SetNullSensorFields();
    RatesInit( g_bInitRates );
    GetCurrentWeatherData( FALSE );

    if( g_HostConfig.nStnType == ID5001 )
        nPca = g_Local.Data.pca;

    WxClear();      // reset all highs/lows to current values; set counts to zero.

    // if stored data is current (i.e., today), load and correct as required
    if( LocalDataStatus() == DATA_OK )
    {
        ReadLocalData();
        // overwrite stored current readings
        if( (g_HostConfig.nStnType==ULT2000) || (g_HostConfig.nStnType==ULTII) )
            while( !UmxDataReady() );     // spin
        GetCurrentWeatherData( FALSE );
    }
    else    // get monthly/annual precip values and override WxClear() on key parms
    {
        ReadLocalPrecipData();
        g_Local.Data.edc = EmpInitTotal();
        g_Local.Data.tyh = 0;
        g_Local.Data.tyl = 0;
    }

    // correct for non-accumlating rain gauges
    if( g_HostConfig.nStnType == ID5001 )
    {
        g_Local.Data.pca = nPca;
        SetLastPrecip( nPca );
        g_Local.Data.pri = 0;
        g_Local.Data.prh = 0;
    }

    SaveLocalData();

    CalculateRiseSet( &g_Local.Info, &g_Local.Data );
    SaveLocalSunData();

    return TRUE;

}


//////////////////////////////////////////////////////////////////////////////
// LocalDataStatus - returns NODATA_ERR, OLDDATA_ERR, or DATA_OK
//

int LocalDataStatus( void )
{
    PGconn * pConn;
    PGresult * pResult;
    char szExec[384];
    int nStatus;

    pConn = PQconnectdb( "dbname=wxn" );
    if( pConn == NULL )
        return NODATA_ERR;
    if( PQstatus( pConn ) != CONNECTION_OK )
    {
        PQfinish( pConn );
        return NODATA_ERR;
    }

    sprintf( szExec, "SELECT dtg_local FROM nodes WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
    pResult = PQexec( pConn, szExec );
    if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
    {
        if( PQgetisnull( pResult, 0, 0 ) )
            nStatus = NODATA_ERR;      // not yet
        else
        {
            PQclear( pResult );
            sprintf( szExec, "SELECT dtg_local FROM nodes WHERE call_ssid='%s' AND CAST( dtg_local AS DATE ) = CURRENT_DATE", g_Local.Info.szCallSsid );
            pResult = PQexec( pConn, szExec );
            if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
                nStatus = DATA_OK;
            else
                nStatus = OLDDATA_ERR;
        }
    }
    else
        nStatus = NODATA_ERR;

    PQclear( pResult );
    PQfinish( pConn );

    return nStatus;
}


//////////////////////////////////////////////////////////////////////////////
// SetOwSensors - If sensors are present, sets appropriate flag.
//

void SetOwSensors( BOOL bVerbose )
{
    if( g_HostConfig.bOwEmp )
    {
        g_Sensors.bOwEmp = TRUE;
        if( bVerbose )
            fprintf( stdout, "EMP sensor enabled\n" );
    }

    if( g_HostConfig.bOwRad )
    {
        if( g_HostConfig.nCesium137 < 10 )
            g_HostConfig.bOwRad = FALSE;
        else
        {
            g_Sensors.bOwRad = TRUE;
            if( bVerbose )
                fprintf( stdout, "Radiation sensor enabled\n" );
        }
    }

    if( g_HostConfig.bOwBarom )
    {
        g_Sensors.bOwBarom = TRUE;
        if( bVerbose )
            fprintf( stdout, "One-wire barometric pressure sensor enabled\n" );
    }

    if( g_HostConfig.bOwSoilTemp )
    {
        g_Sensors.bOwSoilTemp = TRUE;
        if( bVerbose )
            fprintf( stdout, "Soil temperature sensor enabled\n" );
    }

    if( g_HostConfig.bOwSoil2Temp )
    {
        g_Sensors.bOwSoil2Temp = TRUE;
        if( bVerbose )
            fprintf( stdout, "Soil2 temperature sensor enabled\n" );
    }

    if( g_HostConfig.bOwInsideTemp )
    {
        g_Sensors.bOwInsideTemp = TRUE;
        g_Sensors.bInsideTemp = FALSE;      // turn off wx stn sensor, if present
        if( bVerbose )
            fprintf( stdout, "One-wire inside temperature sensor enabled\n" );
    }

    if( g_HostConfig.bOwOutsideTemp )
    {
        g_Sensors.bOwOutsideTemp = TRUE;
        if( bVerbose )
            fprintf( stdout, "One-wire outside temperature sensor enabled\n" );
    }

    if( g_HostConfig.bOwInsideHumid )
    {
        g_Sensors.bOwInsideHumid = TRUE;
        g_Sensors.bInsideHumid = FALSE;     // turn off wx stn sensor, if present
        if( bVerbose )
            fprintf( stdout, "One-wire inside humidity sensor enabled\n" );
    }

    if( g_HostConfig.bOwOutsideHumid )
    {
        g_Sensors.bOwOutsideHumid = TRUE;
        g_Sensors.bOutsideHumid = FALSE;    // turn off wx stn sensor, if present
        if( bVerbose )
            fprintf( stdout, "One-wire outside humidity sensor enabled\n" );
    }

    if( g_HostConfig.bOwSunshine )
    {
        g_Sensors.bOwSun = TRUE;
        if( bVerbose )
            fprintf( stdout, "Sunshine duration sensor interface enabled\n" );
    }

    if( g_HostConfig.bOwAspCtrl )
    {
        g_Sensors.bOwAspCtrl = TRUE;
        if( bVerbose )
            fprintf( stdout, "Aspirator control interface enabled\n" );
    }

    if( g_HostConfig.bOwHtrCtrl )
    {
        g_Sensors.bOwHtrCtrl = TRUE;
        if( bVerbose )
            fprintf( stdout, "Rain gauge heater control interface enabled\n" );
    }
}


//////////////////////////////////////////////////////////////////////////////
// OwSensorsInit - If one-wire sensors are present, performs initialization,
// if required.
//

void OwSensorsInit( void )
{
    int nVal;
    int i;

    if( g_Sensors.bOwEmp )
        EmpInit();

    if( g_Sensors.bOwRad )
        RadInit( g_HostConfig.nCesium137 );

    if( g_Sensors.bOwBarom )
    {
        // try three times in case there is a problem
        i = 0;
        //FOREVER
        for(;;)
        {
            nVal = OwReadBarom();
            if( nVal )
            {
                BaromSet( nVal );
                break;
            }

            if( ++i == 3 )
            {
                g_log.PostMsg( HOSTMSG_OWBAROM_INIT_FAIL, NULL );
                g_Sensors.bOwBarom = FALSE;
                break;
            }
        }
    }

    if( g_Sensors.bOwOutsideHumid )
    {
        if( OwReadHumid( OUTSIDE_HUMIDITY, nVal ) )
            HumidSet( nVal );
        else
        {
            g_log.PostMsg( HOSTMSG_OWHUMID_OUT_INIT_FAIL, NULL );
            g_Sensors.bOwOutsideHumid = FALSE;
        }
    }

    // aspirator control - no init required

    // rain gauge heater control - no init required

    // soil & soil2 temp - no init required

    // inside temp - no init required

    // outside temp - no init requried

    //if( g_Sensors.bSun )
    //    SunTimeInit();

}


//////////////////////////////////////////////////////////////////////////////
// SetNullSensorFields - NULLs fields of off-line sensors. Called only at
// startup.
//

void SetNullSensorFields( void )
{
    PGconn * pConn;
    PGresult * pResult;
    char szExec[384];

    pConn = PQconnectdb( "dbname=wxn" );
    if( pConn == NULL )
        return;
    if( PQstatus( pConn ) != CONNECTION_OK )
    {
        PQfinish( pConn );
        return;
    }

    if( !(g_Sensors.bBarom || g_Sensors.bOwBarom) )
    {
        sprintf( szExec, "SELECT bcr FROM nodes WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
        {
            if( !PQgetisnull( pResult, 0, 0 ) )
            {
                PQclear( pResult );
                sprintf( szExec, "UPDATE nodes SET bcr=NULL, bhr_val=NULL, bhr_dtg=NULL, "
                                 "blr_val=NULL, blr_dtg=NULL, brh=NULL, brd=NULL "
                                 "WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
                pResult = PQexec( pConn, szExec );
            }
        }
        PQclear( pResult );
    }

    if( !(g_Sensors.bInsideTemp || g_Sensors.bOwInsideTemp) )
    {
        sprintf( szExec, "SELECT ist FROM nodes WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
        {
            if( !PQgetisnull( pResult, 0, 0 ) )
            {
                PQclear( pResult );
                sprintf( szExec, "UPDATE nodes SET ist=NULL WHERE call_ssid='%s'",
                    g_Local.Info.szCallSsid );
                pResult = PQexec( pConn, szExec );
            }
        }
        PQclear( pResult );
    }

    if( !(g_Sensors.bInsideHumid || g_Sensors.bOwInsideHumid) )
    {
        sprintf( szExec, "SELECT ish FROM nodes WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) ) {
            if( !PQgetisnull( pResult, 0, 0 ) )
            {
                PQclear( pResult );
                sprintf( szExec, "UPDATE nodes SET ish=NULL WHERE call_ssid='%s'",
                    g_Local.Info.szCallSsid );
                pResult = PQexec( pConn, szExec );
            }
        }
        PQclear( pResult );
    }

    if( !(g_Sensors.bOutsideHumid || g_Sensors.bOwOutsideHumid) )
    {
        sprintf( szExec, "SELECT hcr FROM nodes WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
        {
            if( !PQgetisnull( pResult, 0, 0 ) )
            {
                PQclear( pResult );
                sprintf( szExec, "UPDATE nodes SET hcr=NULL, hhr_val=NULL, hhr_dtg=NULL, "
                                 "hlr_val=NULL, hlr_dtg=NULL, hrh=NULL, hrd=NULL "
                                 "WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
                pResult = PQexec( pConn, szExec );
            }
        }
        PQclear( pResult );
    }

    if( !g_Sensors.bOwEmp )
    {
        sprintf( szExec, "SELECT ecc FROM nodes WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
        {
            if( !PQgetisnull( pResult, 0, 0 ) )
            {
                PQclear( pResult );
                sprintf( szExec, "UPDATE nodes SET ecc=NULL, eqc=NULL, exc=NULL, "
                                 "ehc=NULL, edc=NULL WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
                pResult = PQexec( pConn, szExec );
            }
        }
        PQclear( pResult );
    }

    if( !g_Sensors.bOwRad )
    {
        sprintf( szExec, "SELECT rca FROM nodes WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
        {
            if( !PQgetisnull( pResult, 0, 0 ) )
            {
                PQclear( pResult );
                sprintf( szExec, "UPDATE nodes SET rca=NULL, rha=NULL, rda=NULL, "
                                 "rmax=NULL, ralarm_count=NULL, ralarm_trip=NULL "
                                 "WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
                pResult = PQexec( pConn, szExec );
            }
        }
        PQclear( pResult );
    }

    if( !g_Sensors.bOwSoilTemp )
    {
        sprintf( szExec, "SELECT tsd FROM nodes WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
        {
            if( !PQgetisnull( pResult, 0, 0 ) )
            {
                PQclear( pResult );
                sprintf( szExec, "UPDATE nodes SET tsd=NULL WHERE call_ssid='%s'",
                    g_Local.Info.szCallSsid );
                pResult = PQexec( pConn, szExec );
            }
        }
        PQclear( pResult );
    }

    if( !g_Sensors.bOwSoil2Temp )
    {
        sprintf( szExec, "SELECT tsd2 FROM nodes WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
        {
            if( !PQgetisnull( pResult, 0, 0 ) )
            {
                PQclear( pResult );
                sprintf( szExec, "UPDATE nodes SET tsd2=NULL WHERE call_ssid='%s'",
                    g_Local.Info.szCallSsid );
                pResult = PQexec( pConn, szExec );
            }
        }
        PQclear( pResult );
    }

    if( !g_Sensors.bOwSun )
    {
        sprintf( szExec, "SELECT sun_min FROM nodes WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
        {
            if( !PQgetisnull( pResult, 0, 0 ) )
            {
                PQclear( pResult );
                sprintf( szExec, "UPDATE nodes SET sun_min=NULL, sun_now=NULL WHERE call_ssid='%s'",
                    g_Local.Info.szCallSsid );
                pResult = PQexec( pConn, szExec );
            }
        }
        PQclear( pResult );
    }

    PQfinish( pConn );
}


//////////////////////////////////////////////////////////////////////////////
// ReadLocalPrecipData - reads local data from pgsql. Returns TRUE success,
// FALSE failure.
//

BOOL ReadLocalPrecipData( void )
{
    PGconn * pConn;
    PGresult * pResult;
    char szExec[384];
    BOOL bStatus;

    pConn = PQconnectdb( "dbname=wxn" );
    if( pConn == NULL )
        return FALSE;
    if( PQstatus( pConn ) != CONNECTION_OK )
    {
        PQfinish( pConn );
        return FALSE;
    }

    sprintf( szExec, "SELECT ptm, pta FROM nodes WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
    pResult = PQexec( pConn, szExec );
    if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
    {
        g_Local.Data.ptm = atoi( PQgetvalue( pResult, 0, 0 ) );
        g_Local.Data.pta = atoi( PQgetvalue( pResult, 0, 1 ) );
        bStatus = TRUE;
    }
    else
        bStatus = FALSE;
    PQclear( pResult );

    PQfinish( pConn );

    return bStatus;
}


//////////////////////////////////////////////////////////////////////////////
// ReadLocalData - reads local data from pgsql. Returns -1 if record does
// not exist, 0 on success, and positive error code if a particular type
// cannot be read.
//

int ReadLocalData( void )
{
    PGconn * pConn;
    PGresult * pResult;
    char szExec[384];
    int nStatus;

    pConn = PQconnectdb( "dbname=wxn" );
    if( pConn == NULL )
        return NODATA_ERR;
    if( PQstatus( pConn ) != CONNECTION_OK )
    {
        PQfinish( pConn );
        return NODATA_ERR;
    }

    nStatus = 0;

    // temperature
    sprintf( szExec, "SELECT tcr, thr_val, thr_dtg, tlr_val, tlr_dtg, trd, trh, tyh, tyl "
        "FROM nodes WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
    pResult = PQexec( pConn, szExec );
    if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
    {
        g_Local.Data.tcr = atoi( PQgetvalue( pResult, 0, 0 ) );
        g_Local.Data.thr.val = atoi( PQgetvalue( pResult, 0, 1 ) );
        g_Local.Data.thr.dtg = g_Time.dtg( PQgetvalue( pResult, 0, 2 ) );
        g_Local.Data.tlr.val = atoi( PQgetvalue( pResult, 0, 3 ) );
        g_Local.Data.tlr.dtg = g_Time.dtg( PQgetvalue( pResult, 0, 4 ) );
        g_Local.Data.trd = atoi( PQgetvalue( pResult, 0, 5 ) );
        g_Local.Data.trh = atoi( PQgetvalue( pResult, 0, 6 ) );
        g_Local.Data.tyh = atoi( PQgetvalue( pResult, 0, 7 ) );
        g_Local.Data.tyl = atoi( PQgetvalue( pResult, 0, 8 ) );
    }
    else
        nStatus |= TEMP_ERR;
    PQclear( pResult );

    ////////////// wind
    // current
    sprintf( szExec, "SELECT wgc_spd, wgc_dir, wac_spd, wac_dir "
        "FROM nodes WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
    pResult = PQexec( pConn, szExec );
    if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
    {
        g_Local.Data.wgc.spd = atoi( PQgetvalue( pResult, 0, 0 ) );
        g_Local.Data.wgc.dir = atoi( PQgetvalue( pResult, 0, 1 ) );
        g_Local.Data.wac.spd = atoi( PQgetvalue( pResult, 0, 2 ) );
        g_Local.Data.wac.dir = atoi( PQgetvalue( pResult, 0, 3 ) );
    }
    else
        nStatus |= WIND_ERR;
    PQclear( pResult );

    // high gust
    sprintf( szExec, "SELECT wgh_spd, wgh_dir, wgh_dtg FROM nodes WHERE call_ssid='%s'",
        g_Local.Info.szCallSsid );
    pResult = PQexec( pConn, szExec );
    if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
    {
        g_Local.Data.wgh.spd = atoi( PQgetvalue( pResult, 0, 0 ) );
        g_Local.Data.wgh.dir = atoi( PQgetvalue( pResult, 0, 1 ) );
        g_Local.Data.wgh.dtg = g_Time.dtg( PQgetvalue( pResult, 0, 2 ) );
    }
    else
        nStatus |= WIND_ERR;
    PQclear( pResult );

    // avg hi/lo
    sprintf( szExec, "SELECT wah_spd, wah_dir, wah_dtg, wal_spd, wal_dir, wal_dtg "
        "FROM nodes WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
    pResult = PQexec( pConn, szExec );
    if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
    {
        g_Local.Data.wah.spd = atoi( PQgetvalue( pResult, 0, 0 ) );
        g_Local.Data.wah.dir = atoi( PQgetvalue( pResult, 0, 1 ) );
        g_Local.Data.wah.dtg = g_Time.dtg( PQgetvalue( pResult, 0, 2 ) );
        g_Local.Data.wal.spd = atoi( PQgetvalue( pResult, 0, 3 ) );
        g_Local.Data.wal.dir = atoi( PQgetvalue( pResult, 0, 4 ) );
        g_Local.Data.wal.dtg = g_Time.dtg( PQgetvalue( pResult, 0, 5 ) );
    }
    else
        nStatus |= WIND_ERR;
    PQclear( pResult );

    // wind chill
    sprintf( szExec, "SELECT wca, wcg FROM nodes WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
    pResult = PQexec( pConn, szExec );
    if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
    {
        g_Local.Data.wca = atoi( PQgetvalue( pResult, 0, 0 ) );
        g_Local.Data.wcg = atoi( PQgetvalue( pResult, 0, 1 ) );
    }
    else
        nStatus |= WIND_ERR;
    PQclear( pResult );

    // degree days
    sprintf( szExec, "SELECT cool_days, heat_days FROM nodes WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
    pResult = PQexec( pConn, szExec );
    if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
    {
        g_Local.Data.nCoolDays = atoi( PQgetvalue( pResult, 0, 0 ) );
        g_Local.Data.nHeatDays = atoi( PQgetvalue( pResult, 0, 1 ) );
    }
    else
        nStatus |= DD_ERR;
    PQclear( pResult );

    // precip
    sprintf( szExec, "SELECT pca, pri, prh, plh, p24, pty, ptm, pta "
        "FROM nodes WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
    pResult = PQexec( pConn, szExec );
    if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
    {
        g_Local.Data.pca = atoi( PQgetvalue( pResult, 0, 0 ) );
        g_Local.Data.pri = atoi( PQgetvalue( pResult, 0, 1 ) );
        g_Local.Data.prh = atoi( PQgetvalue( pResult, 0, 2 ) );
        g_Local.Data.plh = atoi( PQgetvalue( pResult, 0, 3 ) );
        g_Local.Data.p24 = atoi( PQgetvalue( pResult, 0, 4 ) );
        g_Local.Data.pty = atoi( PQgetvalue( pResult, 0, 5 ) );
        g_Local.Data.ptm = atoi( PQgetvalue( pResult, 0, 6 ) );
        g_Local.Data.pta = atoi( PQgetvalue( pResult, 0, 7 ) );
    }
    else
        nStatus |= PRECIP_ERR;
    PQclear( pResult );

    // for the following sensors, read only if sensor is attached

    // barometer
    if( g_Sensors.bBarom || g_Sensors.bOwBarom )
    {
        sprintf( szExec, "SELECT bcr, bhr_val, bhr_dtg, blr_val, blr_dtg, brd, brh "
            "FROM nodes WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
        {
            if( !PQgetisnull( pResult, 0, 0 ) )
            {
                g_Local.Data.bcr = atoi( PQgetvalue( pResult, 0, 0 ) );
                g_Local.Data.bhr.val = atoi( PQgetvalue( pResult, 0, 1 ) );
                g_Local.Data.bhr.dtg = g_Time.dtg( PQgetvalue( pResult, 0, 2 ) );
                g_Local.Data.blr.val = atoi( PQgetvalue( pResult, 0, 3 ) );
                g_Local.Data.blr.dtg = g_Time.dtg( PQgetvalue( pResult, 0, 4 ) );
                g_Local.Data.brd = atoi( PQgetvalue( pResult, 0, 5 ) );
                g_Local.Data.brh = atoi( PQgetvalue( pResult, 0, 6 ) );
            }
        }
        else
            nStatus |= BAROM_ERR;
        PQclear( pResult );
    }

    // humidity
    if( g_Sensors.bOutsideHumid || g_Sensors.bOwOutsideHumid )
    {
        sprintf( szExec, "SELECT hcr, hhr_val, hhr_dtg, hlr_val, hlr_dtg, hrd, hrh "
            "FROM nodes WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
        {
            if( !PQgetisnull( pResult, 0, 0 ) )
            {
                g_Local.Data.hcr = atoi( PQgetvalue( pResult, 0, 0 ) );
                g_Local.Data.hhr.val = atoi( PQgetvalue( pResult, 0, 1 ) );
                g_Local.Data.hhr.dtg = g_Time.dtg( PQgetvalue( pResult, 0, 2 ) );
                g_Local.Data.hlr.val = atoi( PQgetvalue( pResult, 0, 3 ) );
                g_Local.Data.hlr.dtg = g_Time.dtg( PQgetvalue( pResult, 0, 4 ) );
                g_Local.Data.hrd = atoi( PQgetvalue( pResult, 0, 5 ) );
                g_Local.Data.hrh = atoi( PQgetvalue( pResult, 0, 6 ) );
            }
        }
        else
            nStatus |= HUMID_ERR;
        PQclear( pResult );
    }

    // emp
    if( g_Sensors.bOwEmp )
    {
        sprintf( szExec, "SELECT ecc, eqc, exc, ehc, edc "
            "FROM nodes WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
        {
            if( !PQgetisnull( pResult, 0, 0 ) )
            {
                g_Local.Data.ecc = (unsigned) atoi( PQgetvalue( pResult, 0, 0 ) );
                g_Local.Data.eqc = (unsigned) atoi( PQgetvalue( pResult, 0, 1 ) );
                g_Local.Data.exc = (unsigned) atoi( PQgetvalue( pResult, 0, 2 ) );
                g_Local.Data.ehc = (unsigned) atoi( PQgetvalue( pResult, 0, 3 ) );
                g_Local.Data.edc = (unsigned) atoi( PQgetvalue( pResult, 0, 4 ) );
            }
        }
        else
            nStatus |= EMP_ERR;
        PQclear( pResult );
    }

    // radiation
    if( g_Sensors.bOwRad )
    {
        sprintf( szExec, "SELECT rca, rha, rda, rmax, ralarm_count "
            "FROM nodes WHERE call_ssid='%s'", g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
        {
            if( !PQgetisnull( pResult, 0, 0 ) )
            {
                g_Local.Data.rca = atoi( PQgetvalue( pResult, 0, 0 ) );
                g_Local.Data.rha = atoi( PQgetvalue( pResult, 0, 1 ) );
                g_Local.Data.rda = atoi( PQgetvalue( pResult, 0, 2 ) );
                g_Local.Data.rmax = atoi( PQgetvalue( pResult, 0, 3 ) );
                g_Local.Data.ralarm_count = atoi( PQgetvalue( pResult, 0, 4 ) );
            }
        }
        else
            nStatus |= RAD_ERR;
        PQclear( pResult );
    }

    // sunshine
    if( g_Sensors.bOwSun )
    {
        sprintf( szExec, "SELECT sun_now, sun_min FROM nodes WHERE call_ssid='%s'",
            g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
        {
            if( !PQgetisnull( pResult, 0, 0 ) )
            {
                g_Local.Data.bSunNow = atoi( PQgetvalue( pResult, 0, 0 ) );
                g_Local.Data.nSunMin = atoi( PQgetvalue( pResult, 0, 1 ) );
            }
        }
        else
            nStatus |= SUN_ERR;
        PQclear( pResult );
    }

    // soil temp
    if( g_Sensors.bOwSoilTemp )
    {
        sprintf( szExec, "SELECT tsd FROM nodes WHERE call_ssid='%s'",
            g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
        {
            if( !PQgetisnull( pResult, 0, 0 ) )
                g_Local.Data.tsd = atoi( PQgetvalue( pResult, 0, 0 ) );
        }
        else
            nStatus |= TSD_ERR;
        PQclear( pResult );
    }

    // soil2 temp
    if( g_Sensors.bOwSoil2Temp )
    {
        sprintf( szExec, "SELECT tsd2 FROM nodes WHERE call_ssid='%s'",
            g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
        {
            if( !PQgetisnull( pResult, 0, 0 ) )
                g_Local.Data.tsd2 = atoi( PQgetvalue( pResult, 0, 0 ) );
        }
        else
            nStatus |= TSD2_ERR;
        PQclear( pResult );
    }

    // inside temp
    if( g_Sensors.bInsideTemp || g_Sensors.bOwInsideTemp )
    {
        sprintf( szExec, "SELECT ist FROM nodes WHERE call_ssid='%s'",
            g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
        {
            if( !PQgetisnull( pResult, 0, 0 ) )
                g_Local.Data.ist = atoi( PQgetvalue( pResult, 0, 0 ) );
        }
        else
            nStatus |= IST_ERR;
        PQclear( pResult );
    }

    // inside humidity
    if( g_Sensors.bInsideHumid || g_Sensors.bOwInsideHumid )
    {
        sprintf( szExec, "SELECT ish FROM nodes WHERE call_ssid='%s'",
            g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
        {
            if( !PQgetisnull( pResult, 0, 0 ) )
                g_Local.Data.ish = atoi( PQgetvalue( pResult, 0, 0 ) );
        }
        else
            nStatus |= ISH_ERR;
        PQclear( pResult );
    }


    PQfinish( pConn );

    return nStatus;
}


//////////////////////////////////////////////////////////////////////////////
// SaveLocalData - saves local data to pgsql.
//

void SaveLocalData( void )
{
    PGconn * pConn;
    PGresult * pResult;
    char szExec[384];
    char szHiTime[32];
    char szLoTime[32];

    pConn = PQconnectdb( "dbname=wxn" );
    if( pConn == NULL )
        return;
    if( PQstatus( pConn ) != CONNECTION_OK )
    {
        PQfinish( pConn );
        return;
    }

    pResult = PQexec( pConn, "BEGIN" );
    PQclear( pResult );

    // current local timestamp in UTC
    g_Time.CurrentLocalTimestamp( szLoTime );
    g_Time.CurrentZuluTimestamp( szHiTime );
    sprintf( szExec, "UPDATE nodes SET dtg_local='%s', dtg_zulu='%s' WHERE call_ssid='%s'",
        szLoTime, szHiTime, g_Local.Info.szCallSsid );
    pResult = PQexec( pConn, szExec );
    PQclear( pResult );

    // temperature
    g_Time.MakeLocalTimestamp( szHiTime, g_Local.Data.thr.dtg );
    g_Time.MakeLocalTimestamp( szLoTime, g_Local.Data.tlr.dtg );
    sprintf( szExec, "UPDATE nodes SET tcr=%i, thr_val=%i, thr_dtg='%s', tlr_val=%i, tlr_dtg='%s', "
        "trd=%i, trh=%i, tyh=%i, tyl=%i WHERE call_ssid='%s'", g_Local.Data.tcr,
        g_Local.Data.thr.val, szHiTime, g_Local.Data.tlr.val, szLoTime,
        g_Local.Data.trd, g_Local.Data.trh, g_Local.Data.tyh, g_Local.Data.tyl,
        g_Local.Info.szCallSsid );
    pResult = PQexec( pConn, szExec );
    PQclear( pResult );

    ////////////// wind
    // current
    sprintf( szExec, "UPDATE nodes SET wgc_spd=%i, wgc_dir=%i, wac_spd=%i, wac_dir=%i "
        "WHERE call_ssid='%s'",
        g_Local.Data.wgc.spd, g_Local.Data.wgc.dir, g_Local.Data.wac.spd, g_Local.Data.wac.dir,
        g_Local.Info.szCallSsid );
    pResult = PQexec( pConn, szExec );
    PQclear( pResult );

    // high gust
    g_Time.MakeLocalTimestamp( szHiTime, g_Local.Data.wgh.dtg );
    sprintf( szExec, "UPDATE nodes SET wgh_spd=%i, wgh_dir=%i, wgh_dtg='%s' WHERE call_ssid='%s'",
        g_Local.Data.wgh.spd, g_Local.Data.wgh.dir, szHiTime, g_Local.Info.szCallSsid );
    pResult = PQexec( pConn, szExec );
    PQclear( pResult );

    // avg hi/lo
    g_Time.MakeLocalTimestamp( szHiTime, g_Local.Data.wah.dtg );
    g_Time.MakeLocalTimestamp( szLoTime, g_Local.Data.wal.dtg );
    sprintf( szExec, "UPDATE nodes SET wah_spd=%i, wah_dir=%i, wah_dtg='%s', "
        "wal_spd=%i, wal_dir=%i, wal_dtg='%s' WHERE call_ssid='%s'",
        g_Local.Data.wah.spd, g_Local.Data.wah.dir, szHiTime,
        g_Local.Data.wal.spd, g_Local.Data.wal.dir, szLoTime,
        g_Local.Info.szCallSsid );
    pResult = PQexec( pConn, szExec );
    PQclear( pResult );

    // wind chill
    if( g_Local.Data.tcr > 50 )
        sprintf( szExec, "UPDATE nodes SET wca=NULL, wcg=NULL WHERE call_ssid='%s'",
            g_Local.Info.szCallSsid );
    else
        sprintf( szExec, "UPDATE nodes SET wca=%i, wcg=%i WHERE call_ssid='%s'",
            g_Local.Data.wca, g_Local.Data.wcg, g_Local.Info.szCallSsid );
    pResult = PQexec( pConn, szExec );
    PQclear( pResult );

    // precip
    sprintf( szExec, "UPDATE nodes SET pca=%i, pri=%i, prh=%i, plh=%i, p24=%i, pty=%i, ptm=%i, pta=%i "
        "WHERE call_ssid='%s'",
        g_Local.Data.pca, g_Local.Data.pri, g_Local.Data.prh, g_Local.Data.plh, g_Local.Data.p24,
        g_Local.Data.pty, g_Local.Data.ptm, g_Local.Data.pta,
        g_Local.Info.szCallSsid );
    pResult = PQexec( pConn, szExec );
    PQclear( pResult );

    // for the following sensors, must write only if sensor is attached

    // barometer
    if( g_Sensors.bBarom || g_Sensors.bOwBarom )
    {
        g_Time.MakeLocalTimestamp( szHiTime, g_Local.Data.bhr.dtg );
        g_Time.MakeLocalTimestamp( szLoTime, g_Local.Data.blr.dtg );
        sprintf( szExec, "UPDATE nodes SET bcr=%i, bhr_val=%i, bhr_dtg='%s', blr_val=%i, blr_dtg='%s', "
            "brd=%i, brh=%i WHERE call_ssid='%s'", g_Local.Data.bcr,
            g_Local.Data.bhr.val, szHiTime, g_Local.Data.blr.val, szLoTime,
            g_Local.Data.brd, g_Local.Data.brh,
            g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        PQclear( pResult );
    }

    // humidity
    if( g_Sensors.bOutsideHumid || g_Sensors.bOwOutsideHumid )
    {
        g_Time.MakeLocalTimestamp( szHiTime, g_Local.Data.hhr.dtg );
        g_Time.MakeLocalTimestamp( szLoTime, g_Local.Data.hlr.dtg );
        sprintf( szExec, "UPDATE nodes SET hcr=%i, hhr_val=%i, hhr_dtg='%s', hlr_val=%i, hlr_dtg='%s', "
            "hrd=%i, hrh=%i WHERE call_ssid='%s'", g_Local.Data.hcr,
            g_Local.Data.hhr.val, szHiTime, g_Local.Data.hlr.val, szLoTime,
            g_Local.Data.hrd, g_Local.Data.hrh,
            g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        PQclear( pResult );

        sprintf( szExec, "UPDATE nodes SET dewpoint=%i WHERE call_ssid='%s'",
            g_Local.Data.nDewPoint, g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        PQclear( pResult );

        if( g_Local.Data.tcr < 68 )
            sprintf( szExec, "UPDATE nodes SET heatindex=NULL WHERE call_ssid='%s'",
                g_Local.Info.szCallSsid );
        else
            sprintf( szExec, "UPDATE nodes SET heatindex=%i WHERE call_ssid='%s'",
                g_Local.Data.nHeatIndex, g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        PQclear( pResult );
    }

    // emp
    if( g_Sensors.bOwEmp )
    {
        sprintf( szExec, "UPDATE nodes SET ecc=%i, eqc=%i, exc=%i, ehc=%i, edc=%i WHERE call_ssid='%s'",
            g_Local.Data.ecc, g_Local.Data.eqc, g_Local.Data.exc, g_Local.Data.ehc, g_Local.Data.edc,
            g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        PQclear( pResult );
    }

    // radiation
    if( g_Sensors.bOwRad )
    {
        sprintf( szExec, "UPDATE nodes SET rca=%i, rha=%i, rda=%i, rmax=%i, ralarm_count=%i, ralarm_trip=%i "
            "WHERE call_ssid='%s'",
            g_Local.Data.rca, g_Local.Data.rha, g_Local.Data.rda, g_Local.Data.rmax,
            g_Local.Data.ralarm_count, g_HostConfig.Alarm.nHiRad,
            g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        PQclear( pResult );
    }

    // sunshine
    if( g_Sensors.bOwSun )
    {
        sprintf( szExec, "UPDATE nodes SET sun_now=%i, sun_min=%i "
            "WHERE call_ssid='%s'",
            g_Local.Data.bSunNow, g_Local.Data.nSunMin,
            g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        PQclear( pResult );
    }

    // soil temp
    if( g_Sensors.bOwSoilTemp )
    {
        sprintf( szExec, "UPDATE nodes SET tsd=%i WHERE call_ssid='%s'",
            g_Local.Data.tsd, g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        PQclear( pResult );
    }

    // soil2 temp
    if( g_Sensors.bOwSoil2Temp )
    {
        sprintf( szExec, "UPDATE nodes SET tsd2=%i WHERE call_ssid='%s'",
            g_Local.Data.tsd2, g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        PQclear( pResult );
    }

    // inside temp
    if( g_Sensors.bInsideTemp || g_Sensors.bOwInsideTemp )
    {
        sprintf( szExec, "UPDATE nodes SET ist=%i WHERE call_ssid='%s'",
            g_Local.Data.ist, g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        PQclear( pResult );
    }

    // inside humidity
    if( g_Sensors.bInsideHumid || g_Sensors.bOwInsideHumid )
    {
        sprintf( szExec, "UPDATE nodes SET ish=%i WHERE call_ssid='%s'",
            g_Local.Data.ish, g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        PQclear( pResult );
    }

    pResult = PQexec( pConn, "COMMIT" );
    PQclear( pResult );

    PQfinish( pConn );
}


//////////////////////////////////////////////////////////////////////////////
// WxStnClear - clear highs/lows/precip, etc. on wx stn (if required)
//

void WxStnClear( void )
{
    // reset weather station, if required

    switch( g_HostConfig.nStnType )
    {
        case ULT2000:
        case ULTII:
            // nothing to do
            break;

        case ID5001:
            Id5Reset();
            break;

        case DAVIS2:
        case DAVISVP:
            // nothing to do
            break;

    }
}


//////////////////////////////////////////////////////////////////////////////
// WxClear - daily clean-up @ midnight.

void WxClear( void )
{
    // temperature
    g_Local.Data.tyh = g_Local.Data.thr.val;
    g_Local.Data.tyl = g_Local.Data.tlr.val;

    g_Local.Data.thr.val = g_Local.Data.tcr;
    g_Local.Data.thr.dtg = g_Time.Qth();

    g_Local.Data.tlr.val = g_Local.Data.tcr;
    g_Local.Data.tlr.dtg = g_Time.Qth();

    // precip
    g_Local.Data.pty = g_Local.Data.pca;
    g_Local.Data.pca = 0;

    // Wind

    g_Local.Data.wgh.spd = g_Local.Data.wgc.spd;
    g_Local.Data.wgh.dir = g_Local.Data.wgc.dir;
    g_Local.Data.wgh.dtg = g_Time.Qth();

    g_Local.Data.wah.spd = g_Local.Data.wac.spd;
    g_Local.Data.wah.dir = g_Local.Data.wac.dir;
    g_Local.Data.wah.dtg = g_Time.Qth();

    g_Local.Data.wal.spd = g_Local.Data.wac.spd;
    g_Local.Data.wal.dir = g_Local.Data.wac.dir;
    g_Local.Data.wal.dtg = g_Time.Qth();


    // humidity

    if( g_Sensors.bOutsideHumid || g_Sensors.bOwOutsideHumid )
    {
        g_Local.Data.hhr.val = g_Local.Data.hcr;
        g_Local.Data.hhr.dtg = g_Time.Qth();
        g_Local.Data.hlr.val = g_Local.Data.hcr;
        g_Local.Data.hlr.dtg = g_Time.Qth();
    }

    // barometer

    if( g_Sensors.bBarom || g_Sensors.bOwBarom )
    {
        g_Local.Data.bhr.val = g_Local.Data.bcr;
        g_Local.Data.bhr.dtg = g_Time.Qth();
        g_Local.Data.blr.val = g_Local.Data.bcr;
        g_Local.Data.blr.dtg = g_Time.Qth();
    }

    // EMP sensor
    if( g_Sensors.bOwEmp )
        EmpClear();

    // Radiation sensor
    if( g_Sensors.bOwRad )
        RadClear();
}


//////////////////////////////////////////////////////////////////////////////
// SaveLocalSunData
//

void SaveLocalSunData( void )
{
    PGconn * pConn;
    PGresult * pResult;
    char szExec[256];
    char szRiseTime[32];
    char szSetTime[32];

    pConn = PQconnectdb( "dbname=wxn" );
    if( pConn == NULL )
        return;
    if( PQstatus( pConn ) != CONNECTION_OK ) {
        PQfinish( pConn );
        return;
        }

    if( g_Local.Data.tSunrise || g_Local.Data.tSunset ) {
        g_Time.MakeRiseSet( g_Local.Data.tSunrise, szRiseTime );
        g_Time.MakeRiseSet( g_Local.Data.tSunset, szSetTime );
        sprintf( szExec, "UPDATE nodes SET sunrise='%s', sunset='%s' WHERE call_ssid='%s'",
            szRiseTime, szSetTime, g_Local.Info.szCallSsid );
        }
    else
        sprintf( szExec, "UPDATE nodes SET sunrise=NULL, sunset=NULL WHERE call_ssid='%s'",
            g_Local.Info.szCallSsid );

    pResult = PQexec( pConn, szExec );
    PQclear( pResult );

    PQfinish( pConn );
}
