/***************************************************************************
 *   Copyright (C) 2005 by John Bennett   *
 *   n4xi@arrl.net   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

//
// msgq.cpp  -  Manage APRS message queue class
//

#define _MSGQ_CPP_

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#include "msgq.h"

MsgQ::MsgQ()
{
    m_nCount = 0;
    m_pHead = NULL;
    m_pTail = NULL;
}

MsgQ::~MsgQ()
{
    qlist * pFree;
    qlist * pList;

    pList = m_pHead;

    while( pList )
    {
        pFree = pList;
        pList = pList->pNext;
        delete pFree;
    }
}


void MsgQ::Init( int nRetries, time_t tTimeout, time_t tDelay )
{
    m_nRetries = nRetries;
    m_tTimeout = tTimeout;
    m_tDelay = tDelay;
}


void MsgQ::Add( char * szFmCall, char * szResp, int nMID, unsigned uCrc )
{
    qlist * pNew;
    char szMID[MAX_MID_SIZE+1];

    pNew = new qlist;
    if( !pNew )
        return;

    sprintf( szMID, "%i", nMID );

    // set pointers
    pNew->pNext = m_pHead;
    pNew->pPrev = NULL;
    if( m_pHead )
        m_pHead->pPrev = pNew;
    else
        m_pTail = pNew;
    m_pHead = pNew;

    // set data
    pNew->uCrc = uCrc;
    strcpy( pNew->szFmCall, szFmCall );
    strcpy( pNew->szResp, szResp );
    strcpy( pNew->szMID, szMID );
    pNew->nRetries = m_nRetries;
    pNew->bDelete = FALSE;
    pNew->tTimeout = time( NULL ) + m_tTimeout;

    m_nCount++;
}


BOOL MsgQ::InList( char * szCallSsid, unsigned uCrc )
{
    qlist * pList;

    pList = m_pHead;

    while( pList )
    {
        if( uCrc == pList->uCrc )
        {
            if( strcmp( szCallSsid, pList->szFmCall ) == 0 )
                return TRUE;
        }
        pList = pList->pNext;
    }

    return FALSE;
}


//////////////////////////////////////////////////////////////////////////////
// Remove - polymorphic functions remove a node from the list. May be called
// with either the message ID or a pointer to the node to be deleted.
//

BOOL MsgQ::Remove( char * szMID )
{
    qlist * pList;

    pList = m_pHead;

    while( pList )
    {
        if( strcmp( pList->szMID, szMID ) == 0 )
        {
            Remove( pList );
            return TRUE;
        }
        pList = pList->pNext;
    }

    return FALSE;
}

void MsgQ::Remove( qlist * pNode )
{
    if( pNode->pPrev )
        (pNode->pPrev)->pNext = pNode->pNext;
    else
        m_pHead = pNode->pNext;     // if previous is NULL, this is the head of the list

    if( pNode->pNext )
        (pNode->pNext)->pPrev = pNode->pPrev;
    else
        m_pTail = pNode->pPrev;

    delete pNode;
    m_nCount--;
}


//////////////////////////////////////////////////////////////////////////////
// RemoveOld - deletes any records that have the bDelete flag set if delay
// time has passed.
//

void MsgQ::RemoveOld( void )
{

    qlist * pList;
    time_t tNow;

    pList = m_pHead;
    tNow = time( NULL );

    while( pList )
    {
        if( pList->bDelete && (tNow > m_tDelay) )
            Remove( pList );
        pList = pList->pNext;
    }

}


//////////////////////////////////////////////////////////////////////////////
// ChkAck - Checks to see if timeout has occurred for any query. If timeout
// occurs, retry count is decremented, timeout value is reset and returns
// pointer to this node in the list. If no timeout has occurred or the list
// is empty, then returns NULL.
//

qlist * MsgQ::ChkAck( void )
{

    qlist * pList;
    time_t tNow;

    if( m_nCount == 0 )
        return NULL;

    tNow = time( NULL );

    // iterate through list
    pList = m_pHead;
    while( pList )
    {
        if( pList->tTimeout < tNow )
        {
            // timeout has occurred
            pList->nRetries--;
            if( pList->nRetries == 0 )
            {
                pList->bDelete = TRUE;
                pList->tTimeout = tNow + m_tDelay;
            }
            else
                pList->tTimeout = tNow + m_tTimeout;
            return pList;
        }
        pList = pList->pNext;
    }

    return NULL;
}


