/***************************************************************************
 *   Copyright (C) 2005 by John Bennett   *
 *   n4xi@arrl.net   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

//
// qserver.cpp  -  query server
//

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <libpq-fe.h>

// libwxn
#include <crc.h>
#include <strxcpy.h>
#include <makeuc.h>
#include <logmsg.h>
#include <ax25.h>
#include <cconfig.h>

#include "qserver.h"
#include "msgq.h"
#include "qglobal.h"


#define ACK_MID_OFFSET  14

#define MAX_CMD_SIZE    3
#define MAX_FIELDS_SIZE 32
#define MAX_HELP_SIZE   48

#define DEFAULT_STYLE   0   // comma delimited
#define HL_STYLE        1
#define WIND_STYLE      2
#define WIND_DTG_STYLE  3

struct querycmd
{
    char szCmd[MAX_CMD_SIZE+1];
    char szFields[MAX_FIELDS_SIZE+1];
    int nFields;
    int nStyle;
    char szHelp[MAX_HELP_SIZE+1];
};

querycmd QCmd[] =
{
    { "t",  "", 0, 0, "Temp CmdSum: t1, t2" },
    { "t1", "", 0, 0, "Cmds1: tc,th,tl,tr,td,ti,ts" },
    { "t2", "", 0, 0, "Cmds2: tt,tb,ca,cg,hx,dh,dc" },
    { "tc", "tcr", 1, DEFAULT_STYLE, "tc=current temp" },
    { "th", "thr_val,thr_dtg", 2, HL_STYLE, "th=today high temp" },
    { "tl", "tlr_val,tlr_dtg", 2, HL_STYLE, "tl=today low temp" },
    { "tr", "trh", 1, DEFAULT_STYLE, "tr=temp rate/hour" },
    { "td", "trd", 1, DEFAULT_STYLE, "td=temp rate/24H" },
    { "ti", "ist", 1, DEFAULT_STYLE, "ti=inside temp" },
    { "ts", "tsd", 1, DEFAULT_STYLE, "ts=soil temp" },
    { "tt", "tyh", 1, DEFAULT_STYLE, "tt=yday high temp" },
    { "tb", "tyl", 1, DEFAULT_STYLE, "tb=yday low temp" },
    { "ca", "wca", 1, DEFAULT_STYLE, "ca=wind chill avg" },
    { "cg", "wcg", 1, DEFAULT_STYLE, "cg=wind chill gust" },
    { "hx", "heatindex", 1, DEFAULT_STYLE, "hx=heat index" },
    { "dh", "heat_days", 1, DEFAULT_STYLE, "dh=heating degree days" },
    { "dc", "cool_days", 1, DEFAULT_STYLE, "dc=cooling degree days" },

    { "w",  "", 0, 0, "Wind CmdSum: w1" },
    { "w1", "", 0, 0, "Cmds: wc,wg,wh,wl,wt" },
    { "wc", "wac_dir,wac_spd", 2, WIND_STYLE, "wc=current wind avg" },
    { "wg", "wgc_dir,wgc_spd", 2, WIND_STYLE, "wg=current wind gust" },
    { "wh", "wah_dir,wah_spd,wah_dtg", 3, WIND_DTG_STYLE, "wh=wind avg high" },
    { "wl", "wal_dir,wal_spd,wal_dtg", 3, WIND_DTG_STYLE, "wl=wind avg low" },
    { "wt", "wgh_dir,wgh_spd,wgh_dtg", 3, WIND_DTG_STYLE, "wt=wind gust high" },

    { "p",  "", 0, 0, "Precip CmdSum: p1" },
    { "p1", "", 0, 0, "Cmds: pc,pi,ph,pl,pd,py,pm,pa" },
    { "pc", "pca", 1, DEFAULT_STYLE, "pc=precip since midnight" },
    { "pi", "pri", 1, DEFAULT_STYLE, "pi=precip rate/inst" },
    { "ph", "prh", 1, DEFAULT_STYLE, "ph=precip rate/hour" },
    { "pl", "plh", 1, DEFAULT_STYLE, "pl=precip last hour" },
    { "pd", "p24", 1, DEFAULT_STYLE, "pd=precip 24 hr amt" },
    { "py", "pty", 1, DEFAULT_STYLE, "py=precip yesterday" },
    { "pm", "ptm", 1, DEFAULT_STYLE, "pm=precip month total" },
    { "pa", "pta", 1, DEFAULT_STYLE, "pa=precip annual total" },

    { "b",  "", 0, 0, "Barom CmdSum: b1" },
    { "b1", "", 0, 0, "Cmds: bc,bh,bl,br,bd" },
    { "bc", "bcr", 1, DEFAULT_STYLE, "bc=current pressure (0.1 mb)" },
    { "bh", "bhr_val,bhr_dtg", 2, HL_STYLE, "bh=high pressure (0.1 mb)" },
    { "bl", "blr_val,blr_dtg", 2, HL_STYLE, "bl=low pressure (0.1 mb)" },
    { "br", "brh", 1, DEFAULT_STYLE, "br=pressure rate/hour (0.1 mb)" },
    { "bd", "brd", 1, DEFAULT_STYLE, "bd=pressure rate/24H (0.1 mb)" },

    { "h",  "", 0, 0, "Humid CmdSum: h1" },
    { "h1", "", 0, 0, "Cmds: hc,hh,hl,hr,hd,hi" },
    { "hc", "hcr", 1, DEFAULT_STYLE, "hc=current humidity" },
    { "hh", "hhr_val,hhr_dtg", 2, HL_STYLE, "hh=high humidity" },
    { "hl", "hlr_val,hlr_dtg", 2, HL_STYLE, "hl=low humidity" },
    { "hr", "hrh", 1, DEFAULT_STYLE, "hr=humidity rate/hr" },
    { "hd", "hrd", 1, DEFAULT_STYLE, "hd=humidity rate/24H" },
    { "hi", "ish", 1, DEFAULT_STYLE, "hi=inside humidity" },

    { "e",  "", 0, 0, "EMP CmdSum: e1" },
    { "e1", "", 0, 0, "Cmds: ec,eq,es,eh,ed,ea" },
    { "ec", "ecc", 1, DEFAULT_STYLE, "ec=EMP 5 min" },
    { "eq", "eqc", 1, DEFAULT_STYLE, "eq=EMP 15 min" },
    { "es", "exc", 1, DEFAULT_STYLE, "es=EMP 30 min" },
    { "eh", "ehc", 1, DEFAULT_STYLE, "eh=EMP 60 min" },
    { "ed", "edc", 1, DEFAULT_STYLE, "ed=EMP day total" },
    { "ea", "ecc,eqc,exc,ehc,edc", 5, DEFAULT_STYLE, "ea=all EMP" },

    { "r",  "", 0, 0, "BkgRad CmdSum: r1" },
    { "r1", "", 0, 0, "Cmds: rc,rh,rd,ra" },
    { "rc", "rca", 1, DEFAULT_STYLE, "rc=bkg rad/avg 5 min" },
    { "rh", "rha", 1, DEFAULT_STYLE, "rh=bkg rad/avg hr" },
    { "rd", "rda", 1, DEFAULT_STYLE, "rd=bkg rad/avg day" },
    { "ra", "rmax,ralarm_count,ralarm_trip", 3, DEFAULT_STYLE, "ra=bkg rad/avg max,#trips,trip pt" },

    { "m",  "", 0, 0, "Misc CmdSum: m1" },
    { "m1", "", 0, 0, "Cmds: fo,dp,sr,na,dz,dl" },
    { "fo", "fog", 1, DEFAULT_STYLE, "fo=fog" },
    { "dp", "dewpoint", 1, DEFAULT_STYLE, "dp=dewpoint" },
    { "sr", "sunrise,sunset", 2, WIND_STYLE, "sr=sunrise/sunset times" },
    { "an", "nws_alert", 1, DEFAULT_STYLE, "an=NWS alert now, 1=yes, 0=no" },
    { "dz", "dtg_zulu", 1, DEFAULT_STYLE, "dz=zulu time last reading" },
    { "dl", "dtg_local", 1, DEFAULT_STYLE, "dl=local time last reading" },

    { "q",  "", 0, 0, "QTH CmdSum: q1" },
    { "q1", "", 0, 0, "Cmds: qt,qh,qe,qn" },
    { "qt", "city,state", 2, DEFAULT_STYLE, "qt=city,state" },
    { "qh", "fips,county,state", 3, DEFAULT_STYLE, "qh=fips,county,state" },
    { "qe", "elevation,haat", 2, DEFAULT_STYLE, "qe=elevation, anemometer HAAT" },
    { "qn", "afos", 1, DEFAULT_STYLE, "qn=NWS AFOS identifier" },

    { "", "", 0, 0, "" }     // empty set signals end of array

};


static int g_nSN = 0;
static char g_szTxCoordinates[32];

static char g_szResults[5][32];


//////////////////////////////////////////////////////////////////////////////
// GetPacket - returns MSG_TYPE, ACK_TYPE, WRU_TYPE or zero if no packet.
//

int GetPacket( char * pPkt )
{
    char szFmCall[9];
    char szFmCallSsid[12];
    char * pS;
    char * pA;
    int nLen;

    *pPkt = 0;

    if( !g_ax25.get() )
        return 0;

    if( g_ax25.FrameType() != U_FRM_UI )
        return 0;

    nLen = g_ax25.data( pPkt, AX25_FRAME_MAX_SIZE );
    if( nLen == 0 )
        return 0;

    *(pPkt+nLen) = 0;  // null terminate the string

    // don't allow self
    g_ax25.FmCall( szFmCall );
    sprintf( szFmCallSsid, "%s-%i", szFmCall, g_ax25.FmSsid() );
    if( strcasecmp( szFmCallSsid, g_QueryConfig.szServerCall ) == 0 )
        return 0;

    // get rid of trailing CR, if it exists
    pS = strchr( pPkt, '\r' );
    if( pS )
       *pS = 0;

    // point past first character for processing both types below
    pS = pPkt+1;

    if( *pPkt == ':' )
    {   // potentially message type
        pA = strchr( pS, ':' );
        if( pA )
        {
            pA++;
            if( strncmp( pA, "ack", 3 ) == 0 )
                return ACK_TYPE;
            else if( strncmp( pA, "rej", 3 ) == 0 )
                return 0;   // ignore rejects - we shouldn't be getting any
            else if( strncmp( pS, "WXN", 2 ) == 0 )
                return WRU_TYPE;
            else
                return MSG_TYPE;
        }
        // else
        //     not a message
    }

    // WRU
    else if( *pPkt == '?' )
    {
        if( strchr( pS, '?' ) )
        {
            if( strncmp( pS, "WXN", 2 ) == 0 )
                return WRU_TYPE;
        }
    }

    return 0;   // not a message or unknown
}


//////////////////////////////////////////////////////////////////////////////
// ParseMsg - checks to see if message is for us. If so, parses into
// message, (optionally) Message ID and returns TRUE. Otherwise, returns
// FALSE.
//
// Example: ':N4XI-3...:tc{001'
//

BOOL ParseMsg( char * pPkt, char * pMsg, char * pMID )
{
    char szMessage[MSG_MAX_SIZE+1];
    char * pC;      // to-call
    char * pS;
    char * pQ;      // query
    char * pI;

    *pMsg = 0;
    *pMID = 0;

    // local working copy
    strcpy( szMessage, pPkt );

    // parse out the call
    pC = szMessage + 1;     // point past first ':'
    pS = strchr( pC, ':' );
    if( !pS )
        return FALSE;
    pQ = pS + 1;    // for later
    pS--;
    while( *pS == ' ' )
        pS--;
    *(pS+1) = 0;

    // is it for us?
    if( strcasecmp( pC, g_QueryConfig.szServerCall ) )
        return FALSE;   // no

    pI = strchr( pQ, '{' );
    if( pI )
    {
        *pI = 0;    // eliminate message ID from query
        pI++;       // point to msg ID
        strcpy( pMID, pI );
    }

    strcpy( pMsg, pQ );

    return TRUE;
}


//////////////////////////////////////////////////////////////////////////////
// SendAck - if message contains message ID, acks the message. Othewise does
// nothing.
//

void SendAck( char * pFmCall, char * pMID )
{
    char szTemp[32];
    int nLen;

    if( *pMID )
    {
        sprintf( szTemp, ":%-9s:ack%s\r", pFmCall, pMID );
        nLen = strlen( szTemp );
        g_ax25.put( g_ComConfig.szAprsToCall, g_ComConfig.szAprsDigis, szTemp, nLen );

        if( g_bVerbose )
            fprintf( stdout, "ACK [%s] sent to %s\n", pMID, pFmCall );
    }
}


//////////////////////////////////////////////////////////////////////////////
// ProcQuery - process incoming query from APRS station. If query has already
// entered queue, returns. Otherwise query is executed and sent back to user.
// In any event, an ACK is sent for the incoming query, regardless if it is
// in the queue or not.
//
// Example: ':N4XI-3...:tc{001'
//

void ProcQuery( char * pPkt )
{
    char szMsg[MAX_MSG_SIZE+1];
    char szFmCall[MSG_CALL_SIZE];
    char szMID[MAX_MID_SIZE+1];
    char szTemp[MAX_MSG_SIZE+1];
    char szResp[MAX_MSG_SIZE+1];
    int nLen;
    unsigned uCrc;

    if( !ParseMsg( pPkt, szMsg, szMID ) )
        return;

    g_ax25.FmCall( szTemp );
    if( g_ax25.FmSsid() )
        sprintf( szFmCall, "%s-%i", szTemp, g_ax25.FmSsid() );
    else
        strcpy( szFmCall, szTemp );

    SendAck( szFmCall, szMID );

    uCrc = Crc( pPkt, strlen( pPkt ) );
    if( g_queue.InList( szFmCall, uCrc ) )
        return;

    if( g_bVerbose )
        fprintf( stdout, "query submitted: [%s] from %s[%s]\n", szMsg, szFmCall, szMID );

    if( *szMsg == 'n' && SameQuery( szMsg, szResp ) )
        sprintf( szTemp, ":%-9s:%s{%i\r", szFmCall, szResp, g_nSN );
    else if( DataQuery( szMsg, szResp, szFmCall ) )
        sprintf( szTemp, ":%-9s:%s{%i\r", szFmCall, szResp, g_nSN );
    else
        sprintf( szTemp, ":%-9s:Err: %s{%i\r", szFmCall, szMsg, g_nSN );
    nLen = strlen( szTemp );
    g_ax25.put( g_ComConfig.szAprsToCall, g_ComConfig.szAprsDigis, szTemp, nLen );

    g_queue.Add( szFmCall, szTemp, g_nSN, uCrc );

    g_nSN = (g_nSN+1) & 0x2ff;  // set next message ID number and wrap to zero as required
}


//////////////////////////////////////////////////////////////////////////////
// ProcAck - checks if ACK is in queue and removes if so
//

void ProcAck( char * pPkt )
{
    char szMID[MAX_MID_SIZE+1];
    char szCall[16];
    BOOL bStatus;

    strxcpy( szMID, pPkt+ACK_MID_OFFSET, MAX_MID_SIZE );
    bStatus = g_queue.Remove( szMID );

    if( g_bVerbose && bStatus )
    {
        g_ax25.FmCall( szCall );
        fprintf( stdout, "ACK [%s] received from %s-%i\n", szMID, szCall, g_ax25.FmSsid() );
    }
}


//////////////////////////////////////////////////////////////////////////////
// TxResponse
//

void TxResponse( qlist * pNode )
{
    int nLen;
    char szTemp[MSG_MAX_SIZE+1];

    sprintf( szTemp, ":%-9s:%s{%s\r", pNode->szFmCall, pNode->szResp, pNode->szMID );
    nLen = strlen( szTemp );
    g_ax25.put( g_ComConfig.szAprsToCall, g_ComConfig.szAprsDigis, szTemp, nLen );

    if( g_bVerbose )
        fprintf( stdout, "Query response retransmitted to: %s\n", pNode->szFmCall );
}


//////////////////////////////////////////////////////////////////////////////
// ProcWRU - posts broadcast of server position
//

void ProcWRU( void )
{
    char szTemp[64];
    int nLen;

    sprintf( szTemp, "!%s>WXN Server\r", g_szTxCoordinates );
    nLen = strlen( szTemp );
    g_ax25.put( g_ComConfig.szAprsToCall, g_ComConfig.szAprsDigis, szTemp, nLen );
}


//////////////////////////////////////////////////////////////////////////////
// SetCoordinates
//

void SetCoordinates( double dLatitude, double dLongitude )
{
    int nDeg;
    int nMin;
    double dTemp1, dTemp2;
    char szTemp[32];
    char cDir;

    dTemp1 = dLatitude;
    nDeg = (int) dTemp1;
    dTemp2 = dTemp1 - (double) nDeg;
    dTemp1 = dTemp2 * 60;

    nMin = (int) dTemp1;
    dTemp2 = dTemp1 - (double) nMin;
    dTemp1 = (double) (nDeg * 100) + (double) nMin +dTemp2;

    if( dTemp1 < 0 )
        cDir = 'S';
    else
        cDir = 'N';

    sprintf( szTemp, "%07.2f%c", dTemp1, cDir );
    strcpy( g_szTxCoordinates, szTemp );

    dTemp1 = dLongitude;
    nDeg = (int) dTemp1;
    dTemp2 = dTemp1 - (double) nDeg;
    dTemp1 = dTemp2 * 60;
    nMin = (int) dTemp1;
    dTemp2 = dTemp1 - (double) nMin;
    dTemp1 = (double) (nDeg * 100) + (double) nMin +dTemp2;

    if( dTemp1 < 0 )
        cDir = 'E';
    else
        cDir = 'W';

    sprintf( szTemp, "/%08.2f%c", dTemp1, cDir );
    strcat( g_szTxCoordinates, szTemp );
}


//////////////////////////////////////////////////////////////////////////////
// DataQuery - send a query to backend or execute other command
//

BOOL DataQuery( char * szCmd, char * szResp, char * pCall )
{
    PGconn * pConn;
    PGresult * pResult;
    char szExec[384];
    char szNodeCall[16];
    querycmd * pQCmd;
    char * pS;
    char * pN;
    char * pCmd;
    BOOL bStatus;
    BOOL bHelp;
    int i;

    *szResp = 0;
    pQCmd = &QCmd[0];

    pS = szCmd;
    while( *pS )
    {
        if( *pS == ' ' || *pS == ',' || *pS == '.' )
        {
            *pS++ = 0;
            if( *pS )
                break;  // pS should point to call-ssid
        }
        else
            pS++;
    }

    if( *pS )
    {
        strcpy( szNodeCall, pS );
        if( strchr( szNodeCall, '-' ) == NULL )
            strcat( szNodeCall, "-0" );
        pN = szNodeCall;
    }
    else
        pN = g_ComConfig.szServerCall;
    makeuc( pN );

    // help request?
    if( *szCmd == '?' )
    {
        pCmd = szCmd+1;
        if( *pCmd == 0 )
        {
            sprintf( szResp, "Cmd Areas: t,w,p,b,h,e,r,q,m" );
            return TRUE;
        }
        bHelp = TRUE;
    }
    else
    {
        bHelp = FALSE;
        pCmd = szCmd;
    }

    while( *pQCmd->szCmd )
    {
        if( strcasecmp( pCmd, pQCmd->szCmd ) == 0 )
            break;
        pQCmd++;
    }

    if( *pQCmd->szCmd == 0 )
        return FALSE;

    pConn = PQconnectdb( "dbname=wxn" );
    if( pConn == NULL )
        return FALSE;
    if( PQstatus( pConn ) != CONNECTION_OK )
    {
        PQfinish( pConn );
        return FALSE;
    }

    //////////////////// record the query
    if( *pS )
        sprintf( szExec, "INSERT INTO queryd( dtg_local,call_ssid,query_cmd,node) VALUES( current_timestamp(0),'%s','%s','%s' )",
            pCall, szCmd, pN );
    else
        sprintf( szExec, "INSERT INTO queryd( dtg_local,call_ssid,query_cmd) VALUES( current_timestamp(0),'%s','%s' )",
                pCall, szCmd );
    pResult = PQexec( pConn, szExec );
    if( PQresultStatus( pResult ) != PGRES_COMMAND_OK )
        g_log.PostMsg( QRYMSG_RECORD_CREATE_FAIL, pCall );
    PQclear( pResult );

    //////////////////////// do backend query

    bStatus = TRUE;     // assume success

    if( pQCmd->nFields && !bHelp )
    {
        sprintf( szExec, "SELECT %s FROM nodes where call_ssid='%s'",
            pQCmd->szFields, pN );
        pResult = PQexec( pConn, szExec );
        if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && (PQntuples( pResult ) > 0) )
        {
            for( i=0; i < pQCmd->nFields; i++ )
                strcpy( g_szResults[i], PQgetvalue( pResult, 0, i ) );
        }
        else
            bStatus = FALSE;
        PQclear( pResult );
    }

    PQfinish( pConn );

    if( !bStatus )
        return FALSE;

    if( bHelp || (pQCmd->nFields == 0) )
        sprintf( szResp, "%s", pQCmd->szHelp );
    else
    {
        //////////////////////// format data from backend
        switch( pQCmd->nFields )
        {
            case 1:
                sprintf( szResp, "%%%s=%s", pQCmd->szCmd,
                    g_szResults[0] );
                break;

            case 2:
                if( pQCmd->nStyle == WIND_STYLE )
                {
                    sprintf( szResp, "%%%s=%s/%s", pQCmd->szCmd,
                        g_szResults[0], g_szResults[1] );
                }
                else    // DEFAULT_STYLE
                    sprintf( szResp, "%%%s=%s,%s", pQCmd->szCmd,
                        g_szResults[0], g_szResults[1] );
                break;

            case 3:
                if( pQCmd->nStyle == WIND_DTG_STYLE )
                {
                    sprintf( szResp, "%%%s=%s/%s@%s", pQCmd->szCmd,
                        g_szResults[0], g_szResults[1], g_szResults[2] );
                }
                else    // DEFAULT_STYLE
                    sprintf( szResp, "%%%s=%s,%s,%s", pQCmd->szCmd,
                        g_szResults[0], g_szResults[1], g_szResults[2] );
                break;

            case 5:
                sprintf( szResp, "%%%s=%s,%s,%s,%s,%s", pQCmd->szCmd,
                    g_szResults[0], g_szResults[1], g_szResults[2], g_szResults[3], g_szResults[4] );
                break;

            default:
                return FALSE;
        }

        if( *pS )
        {
            sprintf( szExec, " (%s)", pS );
            strcat( szResp, szExec );
        }
    }

    return TRUE;
}


//////////////////////////////////////////////////////////////////////////////
// SameQuery - handle SAME data query
//

BOOL SameQuery( char * szCmd, char * szResp )
{
    PGconn * pConn;
    PGresult * pResult;
    char szExec[384];
    char szTemp[128];
    int nTuples;
    int i;
    char * pS;

    pConn = PQconnectdb( "dbname=wxn" );
    if( pConn == NULL )
        return FALSE;
    if( PQstatus( pConn ) != CONNECTION_OK )
    {
        PQfinish( pConn );
        return FALSE;
    }

    // setup
    *szResp = 0;

    // type of event requested
    pS = szCmd + 1;
    if( *pS == 'l' )    // list all current BIDs as '<code>:<bid>' where code is 3 letter event code
    {
        sprintf( szExec, "SELECT code,bid FROM wrevents WHERE current_timestamp < (orig_dtg + duration)" );
        pResult = PQexec( pConn, szExec );
        if( PQresultStatus( pResult ) == PGRES_TUPLES_OK )
        {
            nTuples = PQntuples( pResult );
            if( nTuples )
            {
                for( i=0; i<nTuples; i++ )
                {
                    sprintf( szExec, "%s:%s,", PQgetvalue( pResult, 0, i ), PQgetvalue( pResult, 1, i ) );
                    strcat( szResp, szExec );
                }
            }
            else
                sprintf( szResp, "No current events" );
        }
        PQclear( pResult );
    }

    else if( *pS == 'd' )    // details of event specified by BID
    // <nature>,
    // exp:<time_left>,
    // <county%4>/<state>, ...
    {
        // find argument
        pS++;
        while( *pS )
        {
            if( *pS != ' ' )
                break;
        }
        if( *pS == 0 )
        {
            PQfinish( pConn );
            return FALSE;
        }
        sprintf( szExec, "SELECT wrcodes.nature FROM wrcodes WHERE (SELECT code FROM wrevents WHERE bid=%s) = wrcodes.code",
            pS );
        pResult = PQexec( pConn, szExec );
        if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
        {
            sprintf( szResp, "%s,", PQgetvalue( pResult, 0, 0 ) );
            PQclear( pResult );
            sprintf( szExec, "SELECT select (orig_dtg + duration - current_timestamp) from wrevents where bid=%s",
                pS );
            pResult = PQexec( pConn, szExec );
            if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
            {
                sprintf( szTemp, "%5s", PQgetvalue( pResult, 0, 0 ) );
                if( *szTemp == '-' )
                    strcat( szResp, "expired," );
                else
                {
                    sprintf( szExec, "exp:%s,", szTemp );
                    strcat( szResp, szExec );
                }
                PQclear( pResult );

                sprintf( szExec, "SELECT fips.county,fips.state FROM fips,wrevents,wrevqth "
                    "WHERE wrevents.bid = %s "
                    "AND wrevents.event_id = wrevqth.event_id "
                    "AND wrevqth.fips = fips.fips",
                    pS );
                pResult = PQexec( pConn, szExec );
                if( PQresultStatus( pResult ) == PGRES_TUPLES_OK )
                {
                    nTuples =  PQntuples( pResult );
                    if( nTuples )
                    {
                        for( i=0; i<nTuples; i++ )
                        {
                            sprintf( szExec, "%4s/%s,", PQgetvalue( pResult, 0, i ), PQgetvalue( pResult, 1, i ) );
                            strcat( szResp, szExec );
                        }
                    }
                    else
                        strcpy( szResp, "query error" );
                }
            }
        }
        PQclear( pResult );
    }

    else if( *pS == 'c' )    // list counties for a given bid. counties display first four letters only
    {
        // find argument
        pS++;
        while( *pS )
        {
            if( *pS != ' ' )
                break;
        }
        if( *pS == 0 )
        {
            PQfinish( pConn );
            return FALSE;
        }

        sprintf( szExec, "SELECT fips.county,wrevents.code FROM fips,wrevents,wrevqth "
            "WHERE wrevents.bid = %s "
            "AND wrevents.event_id = wrevqth.event_id "
            "AND wrevqth.fips = fips.fips",
            pS );
        pResult = PQexec( pConn, szExec );
        if( PQresultStatus( pResult ) == PGRES_TUPLES_OK )
        {
            nTuples =  PQntuples( pResult );
            if( nTuples )
            {
                for( i=0; i<nTuples; i++ )
                {
                    if( i == 0 )
                        sprintf( szExec, "%s:%4s,", PQgetvalue( pResult, 0, i ), PQgetvalue( pResult, 1, i ) );
                    else
                        sprintf( szExec, "%4s,", PQgetvalue( pResult, 0, i ) );
                    strcat( szResp, szExec );
                }
            }
            else
                strcpy( szResp, "bid not found" );
        }
        PQclear( pResult );
    }

    else if( *pS=='x' || *pS=='w' || *pS=='a' )
    // 'x' == list all as '<code>:<state>:<bid>'
    // 'w' == list all warnings as '<code>:<state>:<bid>'
    // 'a' == list all watches as '<code>:<state>:<bid>'
    {
        sprintf( szExec, "SELECT DISTINCT wrevents.code,wrevents.bid,fips.state FROM fips,wrevents,wrevqth WHERE "
            "wrevents.event_id = wrevqth.event_id "
            "AND wrevqth.fips = fips.fips "
            "AND current_timestamp < (orig_dtg + duration)" );

        if( *pS == 'w' )
            strcat( szExec, " AND wrevents.code LIKE '__W'" );
        else if( *pS == 'a' )
            strcat( szExec, " AND wrevents.code LIKE '__A'" );

        pResult = PQexec( pConn, szExec );
        if( PQresultStatus( pResult ) == PGRES_TUPLES_OK )
        {
            nTuples = PQntuples( pResult );
            if( nTuples )
            {
                for( i=0; i<nTuples; i++ )
                {
                    sprintf( szExec, "%s:%s:%s,",
                        PQgetvalue( pResult, 0, i ), PQgetvalue( pResult, 1, i ), PQgetvalue( pResult, 2, i ) );
                    strcat( szResp, szExec );
                }
            }
            else
                sprintf( szResp, "no events found" );
        }
        PQclear( pResult );
    }

    else
        sprintf( szResp, "Invalid request " );  // leave trailing space since it will be dropped

    pS = strchr( szResp, 0 ) - 1;
    *pS = 0;

    PQfinish( pConn );

    return TRUE;
}


