/***************************************************************************
 *   Copyright (C) 2005 by John Bennett   *
 *   n4xi@arrl.net   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

//
//  wxnqueryd: query database via APRS messaging
//

#include <stdlib.h>
#include <stdio.h>
#include <syslog.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <netax25/daemon.h>

// libwxn
#include <logmsg.h>
#include <cconfig.h>

#include "qserver.h"
#include "qconfig.h"
#include "msgq.h"


static void signal_handler(int sig);
static void ExitFail( int nMsg, char * sInfo );


// globals

queryconf g_QueryConfig;
Ax25 g_ax25;
MsgQ g_queue;
BOOL g_bVerbose = FALSE;
logmsg g_log;

// locals

static BOOL g_bStop = FALSE;


static void signal_handler(int sig)
{
    switch (sig)
    {
        case SIGTERM:
        case SIGQUIT:
        case SIGHUP:
            g_bStop = TRUE;
            break;

        default:
            break;
    }
}

static void ExitFail( int nMsg, char * sInfo )
{
    g_log.PostMsg( nMsg, sInfo );
    syslog( LOG_ERR, "Terminating - check app log" );
    exit( 1 );
}


int main(int argc, char *argv[])
{
    BOOL bStartDaemon = TRUE;
    timespec ts;
    char szPacket[AX25_FRAME_MAX_SIZE+1];
    qlist * pNode;

    if( argc > 1 )
    {
        if( strcmp( argv[1], "--debug" ) == 0 )
        {
            g_bVerbose = TRUE;
            bStartDaemon = FALSE;
        }
        else if( strcmp(argv[1], "--verbose") == 0 )
            g_bVerbose = TRUE;
        else if( strcmp(argv[1], "--build") == 0 )
        {
            fprintf( stdout, "wxnqueryd: ver %s (built %s @ %s)\n", VERSION, __DATE__, __TIME__ );
            exit( 0 );
        }
    }

    g_log.init( "/var/log/wxn/wxnqueryd.log", g_bVerbose );

    if( bStartDaemon )
    {
        if( !daemon_start(TRUE) )
            ExitFail( MSG_DAEMON_START_FAIL, NULL );
    }

    // set signal handlers
    signal(SIGTERM, signal_handler);
    signal(SIGHUP,  signal_handler);
    signal(SIGQUIT, signal_handler);
    signal(SIGPIPE, SIG_IGN);

    // print signon message
    if( g_bVerbose )
    {
        fprintf( stdout, "wxnqueryd: query daemon for Linux, ver %s (built %s @ %s)\n", VERSION, __DATE__, __TIME__ );
        fprintf( stdout, "Process ID: %i\n", (int) getpid() );
    }

    g_log.PostMsg( MSG_INIT_BEGIN, NULL );
    syslog( LOG_INFO, "started" );

    //////////////////////// read config files
    if( !ReadComConfig() )
        ExitFail( MSG_STARTUP_FAIL, NULL );
    SetCoordinates( g_ComConfig.dLatitude, g_ComConfig.dLongitude );

    if( !ReadQueryConfig() )
        ExitFail( MSG_STARTUP_FAIL, NULL );

    //////////////////////// use call-ssid or alias
    if( g_QueryConfig.bUseAlias )
        strcpy( g_QueryConfig.szServerCall, g_ComConfig.szAlias );
    else
        strcpy( g_QueryConfig.szServerCall, g_ComConfig.szServerCall );

    //////////////////////// ax.25 setup
    g_ax25.setup( g_QueryConfig.szServerCall, g_ComConfig.szAx25Port, FALSE );

    //////////////////////// queue setup
    g_queue.Init( g_QueryConfig.nRetries, (time_t) g_QueryConfig.nFrAckTimer, (time_t) g_QueryConfig.nDelayTime );

    //////////////////////// log the startup as complete
    g_log.PostMsg( MSG_STARTUP_OK, NULL );

    //////////////////////// setup timespec struct for nanosleep
    ts.tv_sec = 0;
    ts.tv_nsec = 200000000; // 0.2 seconds

    //////////////////////// wait for a query
    FOREVER     // for(;;)
    {
        nanosleep( &ts, NULL );

        switch( GetPacket( szPacket ) )
        {
            case MSG_TYPE:
                ProcQuery( szPacket );
                break;

            case ACK_TYPE:
                ProcAck( szPacket );
                break;

            case WRU_TYPE:
                ProcWRU();
                break;

            default:
                break;
        }

        if( g_bStop )
            break;

        do
        {

            pNode = g_queue.ChkAck();
            if( pNode )
                TxResponse( pNode );

            if( g_bStop )
                break;

        } while( pNode );

        // timed delay for removal of old queries
        g_queue.RemoveOld();
    }

    g_log.PostMsg( MSG_DAEMON_STOP, NULL );
    syslog( LOG_INFO, "stopped" );

    return EXIT_SUCCESS;
}

